/*****************************************************************************/
/*
                                 Admin.c

Server administration support functions for HTTPd.  Many of the reports and
forms used for revision are produced by functions within the core-function
module.  The control function allow server restart, exit, abort, mapping file
reload, authorization path reload and authentication cache purging.


VERSION HISTORY
---------------
23-DEC-1999  MGD  modify UNDEFESC reported by DECC v6.2 
30-OCT-1999  MGD  unbundled WATCH functionality to its own module
10-OCT-1999  MGD  show process report,
                  move service statistics report to NET.C module,
                  CacheReport() now only optionally reports cached files
04-SEP-1999  MGD  remove WatchRequestHeader()
11-JUN-1999  MGD  bugfix; WatchFilter()
12-APR-1999  MGD  WatchDataDump() for request and response bodies
10-JAN-1999  MGD  add proxy items to WATCH menu,
                  add proxy report
07-NOV-1998  MGD  WATCH facility
16-MAR-1998  MGD  server abort changed to "exit NOW", added "restart NOW"
05-OCT-1997  MGD  file cache
28-SEP-1997  MGD  request durations
06-SEP-1997  MGD  service list
09-AUG-1997  MGD  message database
07-JUL-1997  MGD  activity report and logging control functions,
                  reworked the control function (it was getting untidy)
08-JUN-1997  MGD  "Other" menu item, with "Clients" and "Scripting" reports,
                  AdminRequestReport() and DclScriptingReport() functions
01-FEB-1997  MGD  new for HTTPd version 4
*/
/*****************************************************************************/

/* standard C header files */
#include <stdio.h>
#include <ctype.h>
#include <string.h>
#include <stdarg.h>

/* VMS related header files */
#include <descrip.h>
#include <jpidef.h>
#include <libdef.h>
#include <libdtdef.h>
#include <prvdef.h>
#include <ssdef.h>
#include <stsdef.h>
#include <syidef.h>

/* application related header files */
#include "wasd.h"

#define WASD_MODULE "ADMIN"

/******************/
/* global storage */
/******************/

char  ErrorLoggingNotEnabled [] = "Logging is not currently enabled.",
      ErrorLoggingAlreadyEnabled [] = "Logging is already enabled.";

/********************/
/* external storage */
/********************/

#ifdef DBUG
extern boolean Debug;
#else
#define Debug 0 
#endif

extern boolean  CacheEnabled,
                ControlExitRequested,
                ControlRestartRequested,
                LoggingEnabled,
                MonitorEnabled,
                OperateWithSysPrv,
                ProtocolHttpsConfigured;

extern int  CacheHits0,
            CacheHits10,
            CacheHits100,
            CacheHits100plus,
            CacheHitCount,
            CacheLoadCount,
            CacheMemoryInUse,
            CacheTotalKBytesMax,
            ConnectCountCurrent,
            ExitStatus,
            ProxyServiceCount,
            RequestHistoryMax,
            WatchDisabled,
            WatchEnabled;

extern unsigned long HttpdProcessId;

extern char  BuildInfo[],
             ErrorSanityCheck[],
             HttpdUserName[],
             ServerHostPort[],
             SoftwareID[],
             TimeGmtString[],
             Utility[];

extern struct AccountingStruct Accounting;
extern struct ConfigStruct  Config;
extern struct MsgStruct  Msgs;
extern struct ProxyAccountingStruct ProxyAccounting;

/*****************************************************************************/
/*
*/ 

AdminBegin
(
struct RequestStruct *rqptr,
void *NextTaskFunction
)
{
   register char  *cptr;

   boolean  UseServerDatabase;
   int  status;
   unsigned short  Length;
   char  FieldName [128],
         FieldValue [256],
         FilePath [256],
         NumberString [16],
         Path [256],
         ProcessIdString [16],
         VirtualHostNamePort [256];

   /*********/
   /* begin */
   /*********/

   if (Debug) fprintf (stdout, "AdminBegin()\n");

   if (rqptr->WatchItem &&
       (WatchEnabled & WATCH_RESPONSE))
      WatchThis (rqptr, FI_LI, WATCH_RESPONSE,
                      "ADMIN !AZ", rqptr->PathInfoPtr);

   if (!rqptr->AccountingDone)
      rqptr->AccountingDone = ++Accounting.DoServerAdminCount;

   if (!rqptr->RemoteUser[0])
   {
      rqptr->ResponseStatusCode = 401;
      ErrorGeneral (rqptr, MsgFor(rqptr,MSG_AUTH_REQUIRED), FI_LI);
      SysDclAst (NextTaskFunction, rqptr);
      return;
   }

   /********************************************/
   /* modules that parse their own query field */
   /********************************************/

   if (strsame (rqptr->PathInfoPtr, INTERNAL_REVISE_HTA,
                sizeof(INTERNAL_REVISE_HTA)-1))
   {
      HTAdminBegin (rqptr, NextTaskFunction);
      return;
   }

   if (strsame (rqptr->PathInfoPtr, INTERNAL_REPORT_ACTIVITY, -1))
   {
      GraphActivityReport (rqptr, NextTaskFunction);
      return;
   }

   if (strsame (rqptr->PathInfoPtr, INTERNAL_GRAPHIC_ACTIVITY, -1))
   {
      GraphActivityPlotBegin (rqptr, NextTaskFunction);
      return;
   }

   if (strsame (rqptr->PathInfoPtr, INTERNAL_REPORT_PROCESS, -1))
   {
      WatchProcessReport (rqptr, NextTaskFunction);
      return;
   }

   if (strsame (rqptr->PathInfoPtr, INTERNAL_REPORT_WATCH, -1))
   {
      if (rqptr->QueryStringLength)
         WatchBegin (rqptr, NextTaskFunction);
      else
         WatchReport (rqptr, NextTaskFunction);
      return;
   }

   if (strsame (rqptr->PathInfoPtr, INTERNAL_CONTROL_DELETE_PROCESS, -1))
   {
      /********************/
      /* delete a process */
      /********************/

      WatchDeleteProcess (rqptr, NextTaskFunction);
      return;
   }

   /**********************/
   /* parse query string */
   /**********************/

   UseServerDatabase = true;
   FilePath[0] = VirtualHostNamePort[0] = Path[0] = ProcessIdString[0] = '\0';

   cptr = rqptr->QueryStringPtr;
   while (*cptr)
   {
      cptr = ParseQueryField (rqptr, cptr,
                              FieldName, sizeof(FieldName),
                              FieldValue, sizeof(FieldValue),
                              FI_LI);
      if (cptr == NULL)
      {
         /* error occured */
         SysDclAst (NextTaskFunction, rqptr);
         return;
      }

      if (strsame (FieldName, "file", -1))
      {
         if (toupper(FieldValue[0]) == 'Y')
            UseServerDatabase = false;
         else
         if (toupper(FieldValue[0]) == 'N')
            UseServerDatabase = true;
      }
      else
      if (strsame (FieldName, "virtual", -1))
         strzcpy (VirtualHostNamePort, FieldValue, sizeof(VirtualHostNamePort));
      else
      if (strsame (FieldName, "server", -1))
      {
         if (toupper(FieldValue[0]) == 'Y')
            UseServerDatabase = true;
         else
         if (toupper(FieldValue[0]) == 'N')
            UseServerDatabase = false;
      }
      else
      if (strsame (FieldName, "edit", -1))
         strzcpy (FilePath, FieldValue, sizeof(FilePath));
      else
      if (strsame (FieldName, "path", -1))
         strzcpy (Path, FieldValue, sizeof(Path));
      else
      if (strsame (FieldName, "pid", -1))
         strzcpy (ProcessIdString, FieldValue, sizeof(ProcessIdString));
      else
      if (strsame (FieldName, "one", -1) ||
          strsame (FieldName, "at", -1))
         strzcpy (NumberString, FieldValue, sizeof(NumberString));
      else
      {
         rqptr->ResponseStatusCode = 400;
         ErrorGeneral (rqptr, "Unknown query field.", FI_LI);
         SysDclAst (NextTaskFunction, rqptr);
         return;
      }
   }

   /***************/
   /* do function */
   /***************/

   if (strsame (rqptr->PathInfoPtr, INTERNAL_ADMIN, -1) ||
       strsame (rqptr->PathInfoPtr, INTERNAL_REPORT, -1) ||
       strsame (rqptr->PathInfoPtr, INTERNAL_REVISE, -1) ||
       strsame (rqptr->PathInfoPtr, INTERNAL_CONTROL, -1))
   {
      /********/
      /* menu */
      /********/

      AdminMenu (rqptr, NextTaskFunction);
      return;
   }
   else
   if (strsame (rqptr->PathInfoPtr, INTERNAL_REPORT,
                sizeof(INTERNAL_REPORT)-1))
   {
      /***********/
      /* reports */
      /***********/

      if (strsame (rqptr->PathInfoPtr, INTERNAL_REPORT_AUTH_USER, -1))
      {
         AuthCacheTreeReport (rqptr, NextTaskFunction);
         return;
      }
      else
      if (strsame (rqptr->PathInfoPtr, INTERNAL_REPORT_CACHE, -1))
      {
         CacheReport (rqptr, NextTaskFunction, false);
         return;
      }
      else
      if (strsame (rqptr->PathInfoPtr, INTERNAL_REPORT_CACHE_ENTRIES, -1))
      {
         CacheReport (rqptr, NextTaskFunction, true);
         return;
      }
      else
      if (strsame (rqptr->PathInfoPtr, INTERNAL_REPORT_CONFIG, -1))
      {
         ConfigReport (rqptr, NextTaskFunction, UseServerDatabase);
         return;
      }
      else
      if (strsame (rqptr->PathInfoPtr, INTERNAL_REPORT_DCL, -1))
      {
         DclScriptingReport (rqptr, NextTaskFunction);
         return;
      }
      else
      if (strsame (rqptr->PathInfoPtr, INTERNAL_REPORT_PROCESS, -1))
      {
         WatchShowProcess (rqptr, NextTaskFunction, ProcessIdString);
         return;
      }
      else
      if (strsame (rqptr->PathInfoPtr, INTERNAL_REPORT_DECNET, -1))
      {
         DECnetScriptingReport (rqptr, NextTaskFunction);
         return;
      }
      else
      if (strsame (rqptr->PathInfoPtr, INTERNAL_REPORT_MEMORY, -1))
      {
         VmReport (rqptr, NextTaskFunction);
         return;
      }
      else
      if (strsame (rqptr->PathInfoPtr, INTERNAL_REPORT_MESSAGES, -1))
      {
         MsgReport (rqptr, NextTaskFunction, UseServerDatabase);
         return;
      }
      else
      if (strsame (rqptr->PathInfoPtr, INTERNAL_REPORT_AUTH_PATHS, -1))
      {
         AuthConfigReport (rqptr, NextTaskFunction,
                           UseServerDatabase, VirtualHostNamePort);
         return;
      }
      else
      if (strsame (rqptr->PathInfoPtr, INTERNAL_REPORT_PEEK, -1))
      {
         WatchPeek (rqptr, NextTaskFunction, NumberString);
         return;
      }
      else
      if (strsame (rqptr->PathInfoPtr, INTERNAL_REPORT_PROXY, -1))
      {
         ProxyMaintReport (rqptr, NextTaskFunction);
         return;
      }
      else
      if (strsame (rqptr->PathInfoPtr, INTERNAL_REPORT_REQUEST, -1))
      {
         RequestReport (rqptr, NextTaskFunction, false);
         return;
      }
      else
      if (strsame (rqptr->PathInfoPtr, INTERNAL_REPORT_REQUEST_HISTORY, -1))
      {
         RequestReport (rqptr, NextTaskFunction, true);
         return;
      }
      else
      if (strsame (rqptr->PathInfoPtr, INTERNAL_REPORT_SHOW_PROCESS, -1))
      {
         WatchShowProcess (rqptr, NextTaskFunction, ProcessIdString);
         return;
      }
      else
      if (strsame (rqptr->PathInfoPtr, INTERNAL_REPORT_STATS, -1))
      {
         AdminReportServerStats (rqptr, NextTaskFunction);
         return;
      }
      else
      if (strsame (rqptr->PathInfoPtr, INTERNAL_REPORT_MAPPING, -1))
      {
         MapUrl_Report (rqptr, NextTaskFunction,
                        UseServerDatabase, VirtualHostNamePort);
         return;
      }
      else
      if (strsame (rqptr->PathInfoPtr, INTERNAL_REPORT_SSL, -1))
      {
         SeSoLaReport (rqptr, NextTaskFunction);
         return;
      }
   }
   else
   if (strsame (rqptr->PathInfoPtr, INTERNAL_REVISE,
                sizeof(INTERNAL_REVISE)-1))
   {
      /**************************/
      /* configuration/revision */
      /**************************/

      if (strsame (rqptr->PathInfoPtr, INTERNAL_REVISE_CONFIG, -1))
      {
         if (FilePath[0])
            UpdEditFileBegin (rqptr);
         else
            ConfigRevise (rqptr, NextTaskFunction, UseServerDatabase);
         return;
      }
      else
      if (strsame (rqptr->PathInfoPtr, INTERNAL_REVISE_HTA, -1))
      {
         AuthConfigReport (rqptr, NextTaskFunction,
                           UseServerDatabase, VirtualHostNamePort);
         return;
      }
   }
   else
   if (strsame (rqptr->PathInfoPtr, INTERNAL_CONTROL,
                sizeof(INTERNAL_CONTROL)-1))
   {
      /******************/
      /* server control */
      /******************/

      AdminControl (rqptr, NextTaskFunction);
      return;
   }

   /********************/
   /* unknown function */
   /********************/

   rqptr->ResponseStatusCode = 403;
   ErrorGeneral (rqptr, "Unknown function.", FI_LI);
   SysDclAst (NextTaskFunction, rqptr);
}

/*****************************************************************************/
/*
Provide a menu of HTTPd server administration functions.
*/ 

AdminMenu
(
struct RequestStruct *rqptr,
void *NextTaskFunction
)
{
   static $DESCRIPTOR (ResponseFaoDsc,
"<HTML>\n\
<HEAD>\n\
!AZ\
<TITLE>HTTPd !AZ ... Server Administration Menu</TITLE>\n\
</HEAD>\n\
<BODY LINK=\"#0000cc\" VLINK=\"#0000cc\">\n\
<H1><NOBR>HTTPd !AZ</NOBR></H1>\n\
<H2>Server Administration Menu</H2>\n\
<P>\n\
\
<TABLE CELLPADDING=5 CELLSPACING=0 BORDER=1>\n\
\
<TR><TH COLSPAN=2>Configuration</TH><TR>\n\
\
<TD COLSPAN=2 VALIGN=top>\n\
\
<TABLE CELLPADDING=3 CELLSPACING=0 BORDER=0>\n\
\
<TR><TD></TD><TD>|</TD>\
<TH COLSPAN=2>Report</TH><TD>|</TD>\n\
<TH COLSPAN=3>Revise</TH><TD>|</TD>\n\
<TH>&nbsp;Action<SUP>*</SUP></TH><TD>|</TD><TR>\n\
\
<TR>\n\
<TH>Statistics</TH>\n\
<TD>&nbsp;</TD>\n\
<TD>[<A HREF=\"!AZ\">Server</A>]</TD>\n\
<TD>&nbsp;</TD>\n\
<TD>&nbsp;</TD>\n\
<TD>&nbsp;</TD>\n\
<TD>&nbsp;</TD>\n\
<TD>&nbsp;</TD>\n\
<TD>&nbsp;</TD>\n\
<TD>[<A HREF=\"!AZ\">Zero</A>]</TD>\n\
</TR>\n\
\
<TR>\n\
<TH>Configuration</TH>\n\
<TD>&nbsp;</TD>\n\
<TD>[<A HREF=\"!AZ?server=yes\">Server</A>]</TD>\n\
<TD>[<A HREF=\"!AZ?file=yes\">File</A>]</TD>\n\
<TD>&nbsp;</TD>\n\
<TD>[<A HREF=\"!AZ?server=yes\">Server</A>]</TD>\n\
<TD>[<A HREF=\"!AZ?file=yes\">File</A>]</TD>\n\
<TD>[<A HREF=\"!AZ!AZ\">Edit</A>]</TD>\n\
</TR>\n\
\
<TR>\n\
<TH>Messages</TH>\n\
<TD>&nbsp;</TD>\n\
<TD>[<A HREF=\"!AZ?server=yes\">Server</A>]</TD>\n\
<TD>[<A HREF=\"!AZ?file=yes\">File</A>]</TD>\n\
<TD>&nbsp;</TD>\n\
<TD>&nbsp;</TD>\n\
<TD>&nbsp;</TD>\n\
<TD>[<A HREF=\"!AZ!AZ\">Edit</A>]</TD>\n\
<TD>&nbsp;</TD>\n\
<TD>&nbsp;</TD>\n\
</TR>\n\
\
<TR>\n\
<TH>Mapping Rules</TH>\n\
<TD>&nbsp;</TD>\n\
<TD>[<A HREF=\"!AZ?server=yes\">Server</A>]</TD>\n\
<TD>[<A HREF=\"!AZ?file=yes\">File</A>]</TD>\n\
<TD>&nbsp;</TD>\n\
<TD>&nbsp;</TD>\n\
<TD>&nbsp;</TD>\n\
<TD>[<A HREF=\"!AZ!AZ\">Edit</A>]</TD>\n\
<TD>&nbsp;</TD>\n\
<TD>[<A HREF=\"!AZ\">Reload</A>]</TD>\n\
</TR>\n\
\
<TR>\n\
<TH>Path Authorization</TH>\n\
<TD>&nbsp;</TD>\n\
<TD>[<A HREF=\"!AZ?server=yes\">Server</A>]</TD>\n\
<TD>[<A HREF=\"!AZ?file=yes\">File</A>]</TD>\n\
<TD>&nbsp;</TD>\n\
<TD>&nbsp;</TD>\n\
<TD>&nbsp;</TD>\n\
<TD>[<A HREF=\"!AZ!AZ\">Edit</A>]</TD>\n\
<TD>&nbsp;</TD>\n\
<TD>[<A HREF=\"!AZ\">Reload</A>]</TD>\n\
</TR>\n\
\
<TR>\n\
<TH>User Authentication</TH>\n\
<TD>&nbsp;</TD>\n\
<TD>[<A HREF=\"!AZ\">Server</A>]</TD>\n\
<TD>&nbsp;</TD>\n\
<TD>&nbsp;</TD>\n\
<TD>&nbsp;</TD>\n\
<TD>[<A HREF=\"!AZ\">HTA</A>]</TD>\n\
<TD>&nbsp;</TD>\n\
<TD>&nbsp;</TD>\n\
<TD>[<A HREF=\"!AZ\">Purge</A>]</TD>\n\
</TR>\n\
\
<TR>\n\
<TH>Other Reports</TH>\n\
<TD>&nbsp;</TD>\n\
<TD COLSPAN=8><NOBR>\n\
[<A HREF=\"!AZ\">Cache</A>]\n\
[<A HREF=\"!AZ\">DCL</A>]\n\
[<A HREF=\"!AZ\">DECnet</A>]\n\
[<A HREF=\"!AZ\">Memory</A>]\n\
[<A HREF=\"!AZ\">Process</A>]\n\
[!AZ!AZ!AZ]\n\
<BR>\n\
[<A HREF=\"!AZ\">Request</A>]\n\
[!AZ!AZ!AZ]\n\
[!AZ!AZ!AZ]\n\
<BR>\n\
[\
!AZ!AZ!AZ\
!AZ!AZ!AZ\
!AZ!AZ!AZ\
!AZ!AZ!AZ\
!AZ!AZ!AZ\
!AZ!AZ!AZ\
!AZ!AZ!AZ\
!AZ!AZ!AZ\
&nbsp;<FONT SIZE=-1>hours activity</FONT>]\n\
</NOBR></TD></TR>\n\
\
</TABLE>\n\
\
<TR>\n\
<TH><FONT COLOR=\"#ff0000\">Control</FONT> <SUP>*</SUP></TH>\n\
\
<TD ROWSPAN=2 ALIGN=CENTER>\n\
\
<TABLE CELLPADDING=3 CELLSPACING=0 BORDER=0>\n\
<TR><TH COLSPAN=4 ALIGN=center><FONT SIZE=-1><NOBR>!AZ</NOBR></FONT></TH></TR>\n\
<TR><TD COLSPAN=4 ALIGN=center><FONT SIZE=-2><NOBR>!AZ (!AZ)</NOBR></FONT></TD></TR>\n\
\
<TR><TH></TH></TR>\n\
\
<TR><TH ROWSPAN=2 ALIGN=left><FONT SIZE=-1>Times</FONT></TH>\n\
<TH ALIGN=right><FONT SIZE=-1>Up:</FONT></TH>\
<TD align=right><FONT SIZE=-1>!AZ</FONT></TD></TR>\n\
\
<TR><TH ALIGN=right><FONT SIZE=-1>CPU:</FONT></TH>\n\
<TD align=right><FONT SIZE=-1>!UL !2ZL:!2ZL:!2ZL.!2ZL</FONT></TD></TR>\n\
\
<TR><TH ROWSPAN=2 ALIGN=left><FONT SIZE=-1>Requests&nbsp;</FONT></TH>\n\
<TH ALIGN=right><FONT SIZE=-1>Read:</FONT></TH>\
<TD align=right><FONT SIZE=-1>!AZ</FONT></TD></TR>\n\
\
<TR><TH ALIGN=right><FONT SIZE=-1>Write:</FONT></TH>\
<TD align=right><FONT SIZE=-1>!AZ</FONT></TD></TR>\n\
\
<TR><TH ROWSPAN=2 ALIGN=left><FONT SIZE=-1>Bytes</FONT></TH>\n\
<TH ALIGN=right><FONT SIZE=-1>Rx:</FONT></TH>\
<TD align=right><FONT SIZE=-1>!AZ</FONT></TD></TR>\n\
\
<TR><TH ALIGN=right><FONT SIZE=-1>Tx:</FONT></TH>\
<TD align=right><FONT SIZE=-1>!AZ</FONT></TD></TR>\n\
\
</TABLE>\n\
\
</TD></TR>\n\
<TR><TD>\n\
\
<TABLE CELLPADDING=3 CELLSPACING=0 BORDER=0>\n\
<TR><TD VALIGN=top ALIGN=left>\n\
\
<TABLE CELLPADDING=3 CELLSPACING=0 BORDER=0>\n\
<TR><TD><A HREF=\"!AZ\">RESTART</A></TD></TR>\n\
<TR><TD><A HREF=\"!AZ\"><FONT SIZE=-1>RESTART</FONT>NOW</A></TD></TR>\n\
<TR><TD><A HREF=\"!AZ\">EXIT</A></TD></TR>\n\
<TR><TD><A HREF=\"!AZ\"><FONT SIZE=-1>EXIT</FONT>NOW</A></TD></TR>\n\
</TABLE>\n\
\
</TD><TD>&nbsp;</TD><TD>&nbsp;</TD><TD>&nbsp;</TD><TD VALIGN=top>\n\
\
<TABLE CELLPADDING=3 CELLSPACING=0 BORDER=0>\n\
<TR><TD ALIGN=right>!AZ</TD><TD>!AZ!AZ!AZ!AZ!AZ</TD></TR>\n\
<TR><TD ALIGN=right>!AZ</TD><TD>!AZ!AZ!AZ!AZ!AZ</TD></TR>\n\
<TR><TD ALIGN=right>!AZ</TD><TD>!AZ!AZ!AZ</TD></TR>\n\
</TABLE>\n\
\
</TD></TR>\n\
\
<TR><TD COLSPAN=5 ALIGN=left>\n\
<SUP>*</SUP>\
<I><FONT COLOR=\"#ff0000\">Caution</FONT>, no confirmation!!</I>\n\
</TD></TR>\n\
\
</TABLE>\n\
\
</TD></TR>\n\
</TABLE>\n\
\
</BODY>\n\
</HTML>\n");

   static unsigned long  JpiCpuTime,
                         Pid;
   static unsigned long  ConnectTime [2],
                         CurrentTime [2],
                         JpiLoginTime [2];

   static char  UpTime [32];

   static $DESCRIPTOR (UpTimeFaoDsc, "!%D\0");
   static $DESCRIPTOR (UpTimeDsc, UpTime);

   static struct
   {
      unsigned short  buf_len;
      unsigned short  item;
      unsigned char   *buf_addr;
      unsigned short  *ret_len;
   }
      JpiItems [] =
   {
      { sizeof(JpiCpuTime), JPI$_CPUTIM, &JpiCpuTime, 0 },
      { sizeof(JpiLoginTime), JPI$_LOGINTIM, &JpiLoginTime, 0 },
      {0,0,0,0}
   };

   register char  *cptr;
   register unsigned long  *vecptr;

   int  status;
   unsigned short  Length;
   unsigned long  FaoVector [128];
   char  Buffer [8192],
         RequestsRead [16],
         RequestsWrite [16],
         BytesRawRx [32],
         BytesRawTx [32];
   $DESCRIPTOR (BufferDsc, Buffer);

   /*********/
   /* begin */
   /*********/

   if (Debug) fprintf (stdout, "AdminMenu()\n");

   if (VMSnok (status = sys$getjpiw (0, &Pid, 0, &JpiItems, 0, 0, 0)))
   {
      rqptr->ErrorTextPtr = "sys$getjpi()";
      ErrorVmsStatus (rqptr, status, FI_LI);
      SysDclAst (NextTaskFunction, rqptr);
      return;
   }

   sys$gettim (&CurrentTime);
   lib$sub_times (&CurrentTime, &JpiLoginTime, &ConnectTime);
   sys$fao (&UpTimeFaoDsc, 0, &UpTimeDsc, &ConnectTime);

   CommaNumber (32,
                Accounting.MethodGetCount +
                Accounting.MethodHeadCount,
                RequestsRead);

   CommaNumber (32,
                Accounting.MethodDeleteCount +
                Accounting.MethodPostCount +
                Accounting.MethodPutCount,
                RequestsWrite);

   CommaNumber (64, &Accounting.QuadBytesRawRx[0], BytesRawRx);
   CommaNumber (64, &Accounting.QuadBytesRawTx[0], BytesRawTx);

   rqptr->ResponsePreExpired = PRE_EXPIRE_ADMIN;
   HTTP_HEADER_200_HTML (rqptr);

   vecptr = FaoVector;

   *vecptr++ = HtmlMetaInfo (rqptr, NULL);
   *vecptr++ = ServerHostPort;
   *vecptr++ = ServerHostPort;

   /* statistics */
   *vecptr++ = INTERNAL_REPORT_STATS;
   *vecptr++ = INTERNAL_CONTROL_ZERO;

   /* configuration */
   *vecptr++ = INTERNAL_REPORT_CONFIG;
   *vecptr++ = INTERNAL_REPORT_CONFIG;
   *vecptr++ = INTERNAL_REVISE_CONFIG;
   *vecptr++ = INTERNAL_REVISE_CONFIG;
   *vecptr++ = INTERNAL_SCRIPT_UPD;
   *vecptr++ = INTERNAL_REVISE_CONFIG;

   /* messages */
   *vecptr++ = INTERNAL_REPORT_MESSAGES;
   *vecptr++ = INTERNAL_REPORT_MESSAGES;
   *vecptr++ = INTERNAL_SCRIPT_UPD;
   *vecptr++ = INTERNAL_REVISE_MESSAGES;

   /* mapping */
   *vecptr++ = INTERNAL_REPORT_MAPPING;
   *vecptr++ = INTERNAL_REPORT_MAPPING;
   *vecptr++ = INTERNAL_SCRIPT_UPD;
   *vecptr++ = INTERNAL_REVISE_MAPPING;
   *vecptr++ = INTERNAL_CONTROL_MAPPING_LOAD;

   /* authorization */
   *vecptr++ = INTERNAL_REPORT_AUTH_PATHS;
   *vecptr++ = INTERNAL_REPORT_AUTH_PATHS;
   *vecptr++ = INTERNAL_SCRIPT_UPD;
   *vecptr++ = INTERNAL_REVISE_AUTH_PATHS;
   *vecptr++ = INTERNAL_CONTROL_AUTH_LOAD;

   /* authentication */
   *vecptr++ = INTERNAL_REPORT_AUTH_USER;
   *vecptr++ = INTERNAL_REVISE_HTA;
   *vecptr++ = INTERNAL_CONTROL_AUTH_PURGE;

   /* other */
   *vecptr++ = INTERNAL_REPORT_CACHE;
   *vecptr++ = INTERNAL_REPORT_DCL;
   *vecptr++ = INTERNAL_REPORT_DECNET;
   *vecptr++ = INTERNAL_REPORT_MEMORY;
   *vecptr++ = INTERNAL_REPORT_PROCESS;

   /* proxy serving */
   if (ProxyServiceCount)
   {
      *vecptr++ = "<A HREF=\"";
      *vecptr++ = INTERNAL_REPORT_PROXY;
      *vecptr++ = "\">Proxy</A>";
   }
   else
   {
      *vecptr++ = "";
      *vecptr++ = "";
      *vecptr++ = "Proxy";
   }

   /* request */
   *vecptr++ = INTERNAL_REPORT_REQUEST;

   /* SSL */
   if (ProtocolHttpsConfigured)
   {
      *vecptr++ = "<A HREF=\"";
      *vecptr++ = INTERNAL_REPORT_SSL;
      *vecptr++ = "\">SSL</A>";
   }
   else
   {
      *vecptr++ = "";
      *vecptr++ = "";
      *vecptr++ = "SSL";
   }

   /* WATCH facility */
   if (WatchDisabled)
   {
      *vecptr++ = "";
      *vecptr++ = "";
      *vecptr++ = "WATCH";
   }
   else
   {
      *vecptr++ = "<A HREF=\"";
      *vecptr++ = INTERNAL_REPORT_WATCH;
      *vecptr++ = "\">WATCH</A>";
   }

   /* activity */
   if (Config.ActivityNumberOfDays)
   {
      *vecptr++ = "<A HREF=\"";
      *vecptr++ = INTERNAL_REPORT_ACTIVITY;
      *vecptr++ = "?1\">1</A>,\n";
      *vecptr++ = "<A HREF=\"";
      *vecptr++ = INTERNAL_REPORT_ACTIVITY;
      *vecptr++ = "?2\">2</A>,\n";
      *vecptr++ = "<A HREF=\"";
      *vecptr++ = INTERNAL_REPORT_ACTIVITY;
      *vecptr++ = "?4\">4</A>,\n";
      *vecptr++ = "<A HREF=\"";
      *vecptr++ = INTERNAL_REPORT_ACTIVITY;
      *vecptr++ = "?12\">12</A>,\n";
      *vecptr++ = "<A HREF=\"";
      *vecptr++ = INTERNAL_REPORT_ACTIVITY;
      *vecptr++ = "?24\">24</A>,\n";
   }
   else
   {
      *vecptr++ = "1,\n";
      *vecptr++ = "";
      *vecptr++ = "";
      *vecptr++ = "2,\n";
      *vecptr++ = "";
      *vecptr++ = "";
      *vecptr++ = "4,\n";
      *vecptr++ = "";
      *vecptr++ = "";
      *vecptr++ = "12,\n";
      *vecptr++ = "";
      *vecptr++ = "";
      *vecptr++ = "24,\n";
      *vecptr++ = "";
      *vecptr++ = "";
   }
   if (Config.ActivityNumberOfDays >= 3)
   {
      *vecptr++ = "<A HREF=\"";
      *vecptr++ = INTERNAL_REPORT_ACTIVITY;
      *vecptr++ = "?72\">72</A>,\n";
   }
   else
   {
      *vecptr++ = "72,\n";
      *vecptr++ = "";
      *vecptr++ = "";
   }
   if (Config.ActivityNumberOfDays >= 7)
   {
      *vecptr++ = "<A HREF=\"";
      *vecptr++ = INTERNAL_REPORT_ACTIVITY;
      *vecptr++ = "?168\">168</A>,\n";
   }
   else
   {
      *vecptr++ = "168,\n";
      *vecptr++ = "";
      *vecptr++ = "";
   }
   if (Config.ActivityNumberOfDays >= 28)
   {
      *vecptr++ = "<A HREF=\"";
      *vecptr++ = INTERNAL_REPORT_ACTIVITY;
      *vecptr++ = "?672\">672</A>";
   }
   else
   {
      *vecptr++ = "672";
      *vecptr++ = "";
      *vecptr++ = "";
   }

   /* informational accounting */
   *vecptr++ = DayDateTime (rqptr->BinaryTime, 20);
   *vecptr++ = rqptr->GmDateTime;
   *vecptr++ = TimeGmtString;
   for (cptr = UpTime; ISLWS(*cptr); cptr++);
   *vecptr++ = (unsigned long)cptr;
   *vecptr++ = JpiCpuTime / 8640000;             /* CPU day */
   *vecptr++ = (JpiCpuTime % 8640000) / 360000;  /* CPU hour */
   *vecptr++ = (JpiCpuTime % 360000) / 6000;     /* CPU minute */
   *vecptr++ = (JpiCpuTime % 6000 ) / 100;       /* CPU second */
   *vecptr++ = JpiCpuTime % 100;                 /* CPU 10 milliseconds */
   *vecptr++ = RequestsRead;
   *vecptr++ = RequestsWrite;
   *vecptr++ = BytesRawRx;
   *vecptr++ = BytesRawTx;

   /* server control menu items */
   *vecptr++ = INTERNAL_CONTROL_RESTART;
   *vecptr++ = INTERNAL_CONTROL_RESTART_NOW;
   *vecptr++ = INTERNAL_CONTROL_EXIT;
   *vecptr++ = INTERNAL_CONTROL_EXIT_NOW;

   /* logging control menu items */
   if (LoggingEnabled)
   {
      *vecptr++ = "<B>Log</B>";
      *vecptr++ = "<NOBR>[On] [<A HREF=\"";
      *vecptr++ = INTERNAL_CONTROL_LOG_OFF;
      *vecptr++ = "\">Off</A>] [<A HREF=\"";
      *vecptr++ = INTERNAL_CONTROL_LOG_FLUSH;
      *vecptr++ = "\">Flush</A>]</NOBR>";
   }
   else
   {
      *vecptr++ = "<B>Log</B>";
      *vecptr++ = "<NOBR>[<A HREF=\"";
      *vecptr++ = INTERNAL_CONTROL_LOG_ON;
      *vecptr++ = "\">On</A>] [Off] [Flush]</NOBR>";
      *vecptr++ = "";
      *vecptr++ = "";
   }

   /* cache control menu items */
   if (CacheEnabled)
   {
      *vecptr++ = "<B>Cache</B>";
      *vecptr++ = "<NOBR>[On] [<A HREF=\"";
      *vecptr++ = INTERNAL_CONTROL_CACHE_OFF;
      *vecptr++ = "\">Off</A>] [<A HREF=\"";
      *vecptr++ = INTERNAL_CONTROL_CACHE_PURGE;
      *vecptr++ = "\">Purge</A>]</NOBR>";
   }
   else
   {
      *vecptr++ = "<B>Cache</B>";
      *vecptr++ = "<NOBR>[<A HREF=\"";
      *vecptr++ = INTERNAL_CONTROL_CACHE_ON;
      *vecptr++ = "\">On</A>] [Off] [Purge]</NOBR>";
      *vecptr++ = "";
      *vecptr++ = "";
   }

   /* proxy control  */
   if (ProxyServiceCount)
   {
      *vecptr++ = "<B>Proxy</B>";
      *vecptr++ = "<NOBR>[<A HREF=\"";
      *vecptr++ = INTERNAL_CONTROL_PROXY_ADJUST;
      *vecptr++ = "\">Adjust</A>]</NOBR>";
   }
   else
   {
      *vecptr++ = "<B>Proxy</B>";
      *vecptr++ = "<NOBR>[Adjust]</NOBR>";
      *vecptr++ = "";
      *vecptr++ = "";
   }

   status = sys$faol (&ResponseFaoDsc, &Length, &BufferDsc, &FaoVector);
   if (VMSnok (status) || status == SS$_BUFFEROVF)
   {
      rqptr->ErrorTextPtr = "sys$faol()";
      ErrorVmsStatus (rqptr, status, FI_LI);
      SysDclAst (NextTaskFunction, rqptr);
      return;
   }

   Buffer[Length] = '\0';
   NetWriteBuffered (rqptr, NextTaskFunction, Buffer, Length);
}

/*****************************************************************************/
/*
Return report on the HTTPd server's activities.
*/ 

AdminReportServerStats
(
struct RequestStruct *rqptr,
void *NextTaskFunction
)
{
/* ALTPRI and SETPRI synonyms? */
#define IMAG_PRIV_0_ALLOWED \
    (~(PRV$M_ALTPRI | PRV$M_PRMMBX | PRV$M_PSWAPM | \
       PRV$M_SYSNAM | PRV$M_SYSPRV))

#define IMAG_PRIV_1_ALLOWED (~(0x00000000))


/*
   VAX VMS does not seem to like "!@UQ" and must be made "!@UJ"!!!
   The VAX version not supporting this doesn't seem to be documented,
   but at run-time returns a %X00000014 status (%SYSTEM-F-BADPARAM).
   Implication: VAX version can only report a maximum of 2^32-1 bytes, or
   4,294,967,295 before overflow, AXP 2^64-1 (calculator gave an ERROR :^)
*/

   static $DESCRIPTOR (ResponseFaoDsc,
"<HTML>\n\
<HEAD>\n\
!AZ\
<TITLE>HTTPd !AZ ... Server Statistics</TITLE>\n\
</HEAD>\n\
<BODY LINK=\"#0000cc\" VLINK=\"#0000cc\">\n\
<H1><NOBR>HTTPd !AZ</NOBR></H1>\n\
<H2>Server Statistics</H2>\n\
!AZ\n\
\
<P><TABLE CELLPADDING=3 CELLSPACING=0 BORDER=1>\n\
<TR><TH>Software</TH></TD></TR>\n\
<TR><TD>\n\
<TABLE CELLPADDING=0 CELLSPACING=5 BORDER=0>\n\
<TR><TH ALIGN=RIGHT>Version:</TH><TD>!AZ</TD></TR>\n\
<TR><TH ALIGN=RIGHT>Build:</TH><TD>!AZ</TD></TR>\n\
</TABLE>\n\
</TD></TR>\n\
</TABLE>\n\
\
<P><TABLE CELLPADDING=3 CELLSPACING=0 BORDER=1>\n\
<TR><TH>Server Process</TH></TD></TR>\n\
<TR><TD>\n\
<TABLE CELLPADDING=0 CELLSPACING=5 BORDER=0>\n\
<TR>\
<TH ALIGN=RIGHT ALIGN=RIGHT>Name:</TH><TD>!AZ</TD>\
<TH ALIGN=RIGHT>PID:</TH><TD><A HREF=\"!AZ?pid=!8XL\">!8XL</A></TD>\
<TH ALIGN=RIGHT>User:</TH><TD>!AZ</TD>\
</TR>\n\
<TR>\
<TH ALIGN=RIGHT>AuthPriv:</TH><TD>!AZ</TD>\
<TH ALIGN=RIGHT>ImagPriv:</TH><TD>!AZ</TD>\
<TH ALIGN=RIGHT>CurPriv:</TH><TD>!AZ</TD>\
</TR>\n\
<TR>\
<TH ALIGN=RIGHT>Startup:</TH><TD>!UL</TD>\
<TH ALIGN=RIGHT>Last Exit:</TH><TD>!AZ</TD>\
<TH ALIGN=RIGHT>Zeroed:</TH><TD>!UL</TD>\
</TR>\n\
<TR>\
<TH ALIGN=RIGHT>Up-time:</TH><TD>!AZ</TD>\
<TH ALIGN=RIGHT>CPU-time:</TH><TD>!UL !2ZL:!2ZL:!2ZL.!2ZL</TD>\n\
</TR>\n\
<TR>\
<TH ALIGN=RIGHT>Pg.Faults:</TH><TD>!UL</TD>\n\
<TH ALIGN=RIGHT>Pg.Used:</TH><TD>!UL%</TD>\
</TR>\n\
<TR>\
<TH ALIGN=RIGHT>WsSize:</TH><TD>!UL</TD>\
<TH ALIGN=RIGHT>WsPeak:</TH><TD>!UL</TD>\n\
<TH ALIGN=RIGHT>PeakVirt:</TH><TD>!UL</TD>\
</TR>\n\
<TR>\
<TH ALIGN=RIGHT>AST:</TH><TD>!UL / !UL</TD>\
<TH ALIGN=RIGHT>BIO:</TH><TD>!UL / !UL</TD>\
<TH ALIGN=RIGHT>BYT:</TH><TD>!UL / !UL</TD>\
</TR>\n\
<TR>\
<TH ALIGN=RIGHT>DIO:</TH><TD>!UL / !UL</TD>\
<TH ALIGN=RIGHT>ENQ:</TH><TD>!UL / !UL</TD>\
<TH ALIGN=RIGHT>FILE:</TH><TD>!UL / !UL</TD>\
</TR>\n\
<TR>\
<TH ALIGN=RIGHT>PRC:</TH><TD>!UL / !UL</TD>\
<TH ALIGN=RIGHT>TQ:</TH><TD>!UL / !UL</TD>\n\
</TR>\n\
</TABLE>\n\
</TD></TR>\n\
</TABLE>\n");

   static $DESCRIPTOR (StatisticsFaoDsc,
"<P><TABLE CELLPADDING=3 CELLSPACING=0 BORDER=1>\n\
<TR><TH>Server Processing</TH></TR>\n\
<TR><TD>\n\
<TABLE CELLPADDING=0 CELLSPACING=0 BORDER=0>\n\
<TR><TD VALIGN=top>\n\
<TABLE CELLPADDING=0 CELLSPACING=5 BORDER=0>\n\
<TR><TH ALIGN=right><U>Connection</U></TH></TR>\n\
<TR><TH ALIGN=RIGHT>Total:</TH><TD>!UL</TD></TR>\n\
<TR><TH ALIGN=RIGHT>Accepted:</TH><TD>!UL</TD></TR>\n\
<TR><TH ALIGN=RIGHT>Rejected:</TH><TD>!UL</TD></TR>\n\
<TR><TH ALIGN=RIGHT>Current:</TH><TD>!UL</TD></TR>\n\
<TR><TH ALIGN=RIGHT>Peak:</TH><TD>!UL</TD></TR>\n\
<TR><TH ALIGN=RIGHT>Busy:</TH><TD>!UL</TD></TR>\n\
<TR><TH ALIGN=RIGHT>SSL:</TH><TD>!UL</TD></TR>\n\
</TABLE>\n\
</TD><TD VALIGN=top>\n\
<TABLE CELLPADDING=0 CELLSPACING=5 BORDER=0>\n\
<TR><TH ALIGN=right><U>Request</U></TH></TR>\n\
<TR><TH ALIGN=right>Total:</TH><TD>!UL</TD></TR>\n\
<TR><TH ALIGN=right>Parsed:</TH><TD>!UL</TD></TR>\n\
<TR><TH ALIGN=right>Error:</TH><TD>!UL</TD></TR>\n\
<TR><TH ALIGN=right>Forbidden:</TH><TD>!UL</TD></TR>\n\
<TR><TH ALIGN=right>Redirect:</TH>\
<TD>!UL / !UL &nbsp;<FONT SIZE=-1><I>(rem/loc)</I></FONT></TD></TR>\n\
<TR><TH ALIGN=right>Keep-Alive:</TH>\
<TD>!UL / !UL &nbsp;<FONT SIZE=-1><I>(total/max)</I></FONT></TD></TR>\n\
<TR><TH ALIGN=right>No-cache:</TH>\
<TD>!UL &nbsp;<FONT SIZE=-1><I>(Pragma:)</I></FONT></TD></TR>\n\
<TR><TH ALIGN=RIGHT>Total Rx:</TH><TD>!AZ</TD></TR>\n\
<TR><TH ALIGN=RIGHT>Tx:</TH><TD>!AZ</TD></TR>\n\
</TABLE>\n\
</TD><TD VALIGN=top>\n\
<TABLE CELLPADDING=0 CELLSPACING=5 BORDER=0>\n\
<TR><TH ALIGN=right><U>Method</U></TH></TR>\n\
<TR><TH ALIGN=right>GET:</TH><TD>!UL</TD></TR>\n\
<TR><TH ALIGN=right>HEAD:</TH><TD>!UL</TD></TR>\n\
<TR><TH ALIGN=right>POST:</TH><TD>!UL</TD></TR>\n\
<TR><TH ALIGN=right>PUT:</TH><TD>!UL</TD></TR>\n\
</TABLE>\n\
</TD><TD VALIGN=top>\n\
<TABLE CELLPADDING=0 CELLSPACING=5 BORDER=0>\n\
<TR><TH ALIGN=right><U>Response</U></TH></TR>\n\
<TR><TH ALIGN=right><I>1nn</I>:</TH><TD>!UL</TD></TR>\n\
<TR><TH ALIGN=right>2nn:</TH><TD>!UL</TD></TR>\n\
<TR><TH ALIGN=right>3nn:</TH><TD>!UL</TD></TR>\n\
<TR><TH ALIGN=right>4nn:</TH><TD>!UL</TD></TR>\n\
<TR><TH ALIGN=right>5nn:</TH><TD>!UL</TD></TR>\n\
<TR><TH ALIGN=right><I>none</I>:</TH><TD>!UL</TD></TR>\n\
</TABLE>\n\
</TD><TD VALIGN=top>\n\
<TABLE CELLPADDING=0 CELLSPACING=5 BORDER=0>\n\
<TR><TH ALIGN=right><U>Duration</U></TH></TR>\n\
<TR><TH ALIGN=right>Min:</TH><TD ALIGN=right>!UL.!#ZL</TD></TR>\n\
<TR><TH ALIGN=right>Max:</TH><TD ALIGN=right>!UL.!#ZL</TD></TR>\n\
<TR><TH ALIGN=right>Ave:</TH><TD ALIGN=right>!UL.!#ZL</TD></TR>\n\
</TABLE>\n\
</TD></TR>\n\
</TABLE>\n\
</TD></TR>\n\
</TABLE>\n\
\
<P><TABLE CELLPADDING=3 CELLSPACING=0 BORDER=1>\n\
<TR><TH>Server Modules</TH></TR>\n\
<TR><TD>\n\
<TABLE CELLPADDING=0 CELLSPACING=0 BORDER=0>\n\
<TR><TD VALIGN=top>\n\
<TABLE CELLPADDING=0 CELLSPACING=5 BORDER=0>\n\
<TR><TH ALIGN=right COLSPAN=2>Admin:</TH><TD>!UL</TD></TR>\n\
<TR><TH ALIGN=right COLSPAN=2>Directory:</TH><TD>!UL</TD></TR>\n\
<TR><TH ALIGN=right COLSPAN=2>File:</TH>\
<TD COLSPAN=2>!UL / !UL &nbsp;<FONT SIZE=-1><I>(all/304)</I></FONT></TD></TR>\n\
<TR><TH ALIGN=right COLSPAN=2>IsMap:</TH><TD>!UL</TD></TR>\n\
<TR><TH ALIGN=right COLSPAN=2>Menu:</TH><TD>!UL</TD></TR>\n\
<TR><TH ALIGN=right COLSPAN=2>Put:</TH><TD>!UL</TD></TR>\n\
<TR><TH ALIGN=right COLSPAN=2>Proxy:</TH><TD COLSPAN=2>!UL</TD></TR>\n\
<TR><TH ALIGN=right COLSPAN=2>SSI:</TH><TD>!UL</TD></TR>\n\
<TR><TH ALIGN=right COLSPAN=2>Update:</TH><TD>!UL</TD></TR>\n\
</TABLE>\n\
</TD><TD VALIGN=top>\n\
<TABLE CELLPADDING=0 CELLSPACING=5 BORDER=0>\n\
<TR><TH ALIGN=right><U>DCL</U></TH></TR>\n\
<TR><TH ALIGN=right>CGI-script:</TH><TD>!UL</TD></TR>\n\
<TR><TH ALIGN=right>CGIplus-script:</TH>\
<TD>!UL / !UL &nbsp;<FONT SIZE=-1><I>(all/reused)</I></FONT></TD></TR>\n\
<TR><TH ALIGN=right>Autoscripted:</TH><TD>!UL</TD></TR>\n\
<TR><TH ALIGN=right>CLI:</TH><TD>!UL</TD></TR>\n\
<TR><TH ALIGN=right>Spawned:</TH><TD>!UL</TD></TR>\n\
</TABLE>\n\
</TD><TD VALIGN=top>\n\
<TABLE CELLPADDING=0 CELLSPACING=5 BORDER=0>\n\
<TR><TH ALIGN=right><U>DECnet</U></TH></TR>\n\
<TR><TH ALIGN=right>Connect:</TH>\
<TD>!UL / !UL &nbsp;<FONT SIZE=-1><I>(all/fail)</I></FONT></TD></TR>\n\
<TR><TH ALIGN=right>CGI-script:</TH><TD>!UL</TD></TR>\n\
<TR><TH ALIGN=right>OSU-script:</TH><TD>!UL</TD></TR>\n\
</TABLE>\n\
</TD></TR>\n\
</TABLE>\n\
</TD></TR>\n\
</TABLE>\n\
\
<P><TABLE CELLPADDING=3 CELLSPACING=0 BORDER=1>\n\
<TR><TH>File Cache</TH></TR>\n\
<TR><TD>\n\
<TABLE CELLPADDING=0 CELLSPACING=5 BORDER=0>\n\
<TR><TH ALIGN=right>Memory:</TH>\
<TD COLSPAN=2>!UL / !UL kB\
 &nbsp;<FONT SIZE=-1><I>(cur/max)</I></FONT></TD></TR>\n\
<TR><TH ALIGN=right>Loads:</TH><TD>!UL</TD></TR>\n\
<TR><TH ALIGN=right>Not Hit:</TH><TD>!UL</TD><TD>!UL%</TD></TR>\n\
<TR><TH ALIGN=right>Total Hit:</TH><TD>!UL</TD><TD>!UL%</TD></TR>\n\
<TR><TH ALIGN=right>1-9:</TH><TD>!UL</TD><TD>!UL%</TD></TR>\n\
<TR><TH ALIGN=right>10-99:</TH><TD>!UL</TD><TD>!UL%</TD></TR>\n\
<TR><TH ALIGN=right>&gt;100:</TH><TD>!UL</TD><TD>!UL%</TD></TR>\n\
<TR><TD COLSPAN=3><FONT SIZE=-1>\
<SUP>*</SUP><I>counts are per-startup only</I></FONT></TD></TR>\n\
</TABLE>\n\
</TD></TR>\n\
</TABLE>\n\
\
<P><TABLE CELLPADDING=3 CELLSPACING=0 BORDER=1>\n\
<TR><TH>Authorization</TH></TR>\n\
<TR><TD>\n\
<TABLE CELLPADDING=0 CELLSPACING=0 BORDER=0>\n\
<TR><TH><U>Paths</U></TH><TH COLSPAN=2><U>Authentication</U></TH></TR>\n\
<TR><TD VALIGN=top>\n\
<TABLE CELLPADDING=0 CELLSPACING=5 BORDER=0>\n\
<TR><TH ALIGN=right COLSPAN=2>Success:</TH><TD>!UL</TD></TR>\n\
<TR><TH ALIGN=right COLSPAN=2>Failure:</TH><TD>!UL</TD></TR>\n\
</TABLE>\n\
</TD><TD VALIGN=top>\n\
<TABLE CELLPADDING=0 CELLSPACING=5 BORDER=0>\n\
<TR><TH ALIGN=right><U>Source</U></TH>\n\
<TR><TH ALIGN=right>Cache:</TH><TD>!UL</TD>\n\
<TR><TH ALIGN=right>Agent:</TH><TD >!UL</TD>\n\
<TR><TH ALIGN=right>HTA:</TH><TD >!UL</TD>\n\
<TR><TH ALIGN=right>List:</TH><TD >!UL</TD>\n\
<TR><TH ALIGN=right>SYSUAF:</TH><TD>!UL</TD></TR>\n\
</TABLE>\n\
</TD><TD VALIGN=top>\n\
<TABLE CELLPADDING=0 CELLSPACING=5 BORDER=0>\n\
<TR><TH ALIGN=right><U>Scheme</U></TH></TR>\n\
<TR><TH ALIGN=right>Basic:</TH><TD>!UL</TD></TR>\n\
<TR><TH ALIGN=right>Digest:</TH><TD>!UL</TD></TR>\n\
</TABLE>\n\
</TD></TR>\n\
</TABLE>\n\
</TD></TR>\n\
</TABLE>\n");

   static $DESCRIPTOR (OtherFaoDsc,
"<P><TABLE CELLPADDING=3 CELLSPACING=0 BORDER=1>\n\
<TR><TH>Other</TH></TR>\n\
<TR><TD>\n\
<TABLE CELLPADDING=0 CELLSPACING=5 BORDER=0>\n\
<TR><TH ALIGN=right>StreamLF document conversions:</TH><TD>!UL</TD></TR>\n\
</TABLE>\n\
</TD></TR>\n\
</TABLE>\n\
\
</BODY>\n\
</HTML>\n");

   static char  SysPrvWarning[] = "<FONT COLOR=\"#ff0000\">SYSPRV</FONT>";
   static char  PrivWarning[] = "<FONT COLOR=\"#ff0000\">WARNING</FONT>";
   static char  PrivExpected[] = "as expected";

   static char  JpiPrcNam [16],
                JpiUserName [13],
                LastExitStatus [16],
                UpTime [32];

   static unsigned long  JpiAstCnt,
                         JpiAstLm,
                         JpiBioCnt,
                         JpiBytLm,
                         JpiBytCnt,
                         JpiBioLm,
                         JpiCpuTime,
                         JpiDioCnt,
                         JpiDioLm,
                         JpiEnqCnt,
                         JpiEnqLm,
                         JpiFilCnt,
                         JpiFilLm,
                         JpiPageFlts,
                         JpiPagFilCnt,
                         JpiPgFlQuota,
                         JpiPid,
                         JpiPrcCnt,
                         JpiPrcLm,
                         JpiTqCnt,
                         JpiTqLm,
                         JpiVirtPeak,
                         JpiWsSize,
                         JpiWsPeak,
                         Pid;

   static unsigned long  ConnectTime [2],
                         CurrentTime [2],
                         JpiAuthPriv [2],
                         JpiCurPriv [2],
                         JpiImagPriv [2],
                         JpiLoginTime [2];

   static $DESCRIPTOR (LastExitStatusDsc, LastExitStatus);
   static $DESCRIPTOR (LastExitStatusFaoDsc, "%X!XL\0");
   static $DESCRIPTOR (UpTimeFaoDsc, "!%D\0");
   static $DESCRIPTOR (UpTimeDsc, UpTime);
   static struct
   {
      unsigned short  buf_len;
      unsigned short  item;
      unsigned char   *buf_addr;
      unsigned short  *ret_len;
   }
      JpiItems [] =
   {
      { sizeof(JpiAstCnt), JPI$_ASTCNT, &JpiAstCnt, 0 },
      { sizeof(JpiAstLm), JPI$_ASTLM, &JpiAstLm, 0 },
      { sizeof(JpiAuthPriv), JPI$_AUTHPRIV, &JpiAuthPriv, 0 },
      { sizeof(JpiBioCnt), JPI$_BIOCNT, &JpiBioCnt, 0 },
      { sizeof(JpiBioLm), JPI$_BIOLM, &JpiBioLm, 0 },
      { sizeof(JpiBytCnt), JPI$_BYTCNT, &JpiBytCnt, 0 },
      { sizeof(JpiBytLm), JPI$_BYTLM, &JpiBytLm, 0 },
      { sizeof(JpiCpuTime), JPI$_CPUTIM, &JpiCpuTime, 0 },
      { sizeof(JpiCurPriv), JPI$_CURPRIV, &JpiCurPriv, 0 },
      { sizeof(JpiDioCnt), JPI$_DIOCNT, &JpiDioCnt, 0 },
      { sizeof(JpiDioLm), JPI$_DIOLM, &JpiDioLm, 0 },
      { sizeof(JpiEnqCnt), JPI$_ENQCNT, &JpiEnqCnt, 0 },
      { sizeof(JpiEnqLm), JPI$_ENQLM, &JpiEnqLm, 0 },
      { sizeof(JpiFilCnt), JPI$_FILCNT, &JpiFilCnt, 0 },
      { sizeof(JpiFilLm), JPI$_FILLM, &JpiFilLm, 0 },
      { sizeof(JpiImagPriv), JPI$_IMAGPRIV, &JpiImagPriv, 0 },
      { sizeof(JpiLoginTime), JPI$_LOGINTIM, &JpiLoginTime, 0 },
      { sizeof(JpiPageFlts), JPI$_PAGEFLTS, &JpiPageFlts, 0 },
      { sizeof(JpiPagFilCnt), JPI$_PAGFILCNT, &JpiPagFilCnt, 0 },
      { sizeof(JpiPgFlQuota), JPI$_PGFLQUOTA, &JpiPgFlQuota, 0 },
      { sizeof(JpiPid), JPI$_PID, &JpiPid, 0 },
      { sizeof(JpiPrcCnt), JPI$_PRCCNT, &JpiPrcCnt, 0 },
      { sizeof(JpiPrcLm), JPI$_PRCLM, &JpiPrcLm, 0 },
      { sizeof(JpiPrcNam), JPI$_PRCNAM, &JpiPrcNam, 0 },
      { sizeof(JpiTqCnt), JPI$_TQCNT, &JpiTqCnt, 0 },
      { sizeof(JpiTqLm), JPI$_TQLM, &JpiTqLm, 0 },
      { sizeof(JpiUserName), JPI$_USERNAME, &JpiUserName, 0 },
      { sizeof(JpiVirtPeak), JPI$_VIRTPEAK, &JpiVirtPeak, 0 },
      { sizeof(JpiWsSize), JPI$_WSSIZE, &JpiWsSize, 0 },
      { sizeof(JpiWsPeak), JPI$_WSPEAK, &JpiWsPeak, 0 },
      {0,0,0,0}
   };

   register unsigned int  bit;
   register unsigned long  *vecptr;
   register char  *cptr;

   int  status;

   unsigned short  Length;

   unsigned long  ResponseDuration,
                  Remainder;

   unsigned long  FaoVector [128],
                  QuadBytesRawRx [2],
                  QuadBytesRawTx [2];

   char  Buffer [8192],
         BytesRawRx [32],
         BytesRawTx [32];

   char  *AuthPrivPtr,
         *CurPrivPtr,
         *ImagPrivPtr;

   $DESCRIPTOR (BufferDsc, Buffer);

   /*********/
   /* begin */
   /*********/

   if (Debug) fprintf (stdout, "AdminReportServerStats()\n");

   if (VMSnok (status = sys$getjpiw (0, &Pid, 0, &JpiItems, 0, 0, 0)))
   {
      rqptr->ErrorTextPtr = "sys$getjpi()";
      ErrorVmsStatus (rqptr, status, FI_LI);
      SysDclAst (NextTaskFunction, rqptr);
      return;
   }

   JpiUserName[12] = '\0';
   for (cptr = JpiUserName; *cptr && *cptr != ' '; cptr++);
   *cptr = '\0';
   if (Debug) fprintf (stdout, "JpiUserName |%s|\n", JpiUserName);

   JpiPrcNam[15] = '\0';
   for (cptr = JpiPrcNam; *cptr && *cptr != ' '; cptr++);
   *cptr = '\0';
   if (Debug) fprintf (stdout, "JpiPrcNam |%s|\n", JpiPrcNam);

   sys$gettim (&CurrentTime);
   lib$sub_times (&CurrentTime, &JpiLoginTime, &ConnectTime);
   sys$fao (&UpTimeFaoDsc, 0, &UpTimeDsc, &ConnectTime);

   if (Accounting.LastExitStatus)
      sys$fao (&LastExitStatusFaoDsc, 0, &LastExitStatusDsc, 
               Accounting.LastExitStatus);
   else
      strcpy (LastExitStatus, "<I>(none)</I>");

   if ((JpiAuthPriv[0] & ~(PRV$M_NETMBX | PRV$M_TMPMBX)) || JpiAuthPriv[1])
      AuthPrivPtr = PrivWarning;
   else
      AuthPrivPtr = PrivExpected;

   if ((JpiImagPriv[0] & IMAG_PRIV_0_ALLOWED) ||
       (JpiImagPriv[1] & IMAG_PRIV_1_ALLOWED))
      ImagPrivPtr = PrivWarning;
   else
      ImagPrivPtr = PrivExpected;

   if (OperateWithSysPrv)
   {
      if ((JpiCurPriv[0] & ~(PRV$M_NETMBX | PRV$M_TMPMBX | PRV$M_SYSPRV)) ||
          JpiCurPriv[1])
         CurPrivPtr = PrivWarning;
      else
         CurPrivPtr = SysPrvWarning;
   }
   else
   if ((JpiCurPriv[0] & ~(PRV$M_NETMBX | PRV$M_TMPMBX)) || JpiCurPriv[1])
      CurPrivPtr = PrivWarning;
   else
      CurPrivPtr = PrivExpected;

   rqptr->ResponsePreExpired = PRE_EXPIRE_ADMIN;
   HTTP_HEADER_200_HTML (rqptr);

   vecptr = FaoVector;

   /* META info */
   *vecptr++ = HtmlMetaInfo (rqptr, NULL);

   /* page title, etc. */
   *vecptr++ = ServerHostPort;
   *vecptr++ = ServerHostPort;
   *vecptr++ = DayDateTime (&rqptr->BinaryTime, 20);

   /* server */
   *vecptr++ = (unsigned long)SoftwareID;
   *vecptr++ = (unsigned long)BuildInfo;
   *vecptr++ = (unsigned long)JpiPrcNam;
   *vecptr++ = INTERNAL_REPORT_SHOW_PROCESS;
   *vecptr++ = JpiPid;
   *vecptr++ = JpiPid;
   *vecptr++ = (unsigned long)JpiUserName;
   *vecptr++ = AuthPrivPtr;
   *vecptr++ = ImagPrivPtr;
   *vecptr++ = CurPrivPtr;
   *vecptr++ = Accounting.StartupCount;
   *vecptr++ = LastExitStatus;
   *vecptr++ = Accounting.ZeroedCount;
   for (cptr = UpTime; ISLWS(*cptr); cptr++);
   *vecptr++ = (unsigned long)cptr;
   *vecptr++ = JpiCpuTime / 8640000;             /* CPU day */
   *vecptr++ = (JpiCpuTime % 8640000) / 360000;  /* CPU hour */
   *vecptr++ = (JpiCpuTime % 360000) / 6000;     /* CPU minute */
   *vecptr++ = (JpiCpuTime % 6000 ) / 100;       /* CPU second */
   *vecptr++ = JpiCpuTime % 100;                 /* CPU 10 milliseconds */
   *vecptr++ = JpiPageFlts;
   *vecptr++ = 100 - ((JpiPagFilCnt * 100) / JpiPgFlQuota);
   *vecptr++ = JpiWsSize;
   *vecptr++ = JpiWsPeak;
   *vecptr++ = JpiVirtPeak;
   *vecptr++ = JpiAstLm - JpiAstCnt;
   *vecptr++ = JpiAstLm;
   *vecptr++ = JpiBioLm - JpiBioCnt;
   *vecptr++ = JpiBioLm;
   *vecptr++ = JpiBytLm - JpiBytCnt;
   *vecptr++ = JpiBytLm;
   *vecptr++ = JpiDioLm - JpiDioCnt;
   *vecptr++ = JpiDioLm;
   *vecptr++ = JpiEnqLm - JpiEnqCnt;
   *vecptr++ = JpiEnqLm;
   *vecptr++ = JpiFilLm - JpiFilCnt;
   *vecptr++ = JpiFilLm;
   *vecptr++ = JpiPrcCnt;
   *vecptr++ = JpiPrcLm;
   *vecptr++ = JpiTqLm - JpiTqCnt;
   *vecptr++ = JpiTqLm;

   status = sys$faol (&ResponseFaoDsc, &Length, &BufferDsc, &FaoVector);
   if (VMSnok (status) || status == SS$_BUFFEROVF)
   {
      rqptr->ErrorTextPtr = "sys$faol()";
      ErrorVmsStatus (rqptr, status, FI_LI);
      SysDclAst (NextTaskFunction, rqptr);
      return;
   }

   Buffer[Length] = '\0';
   NetWriteBuffered (rqptr, 0, Buffer, Length);

   /************/
   /* services */
   /************/

   if (VMSnok (NetServiceReportStats (rqptr)))
   {
      SysDclAst (NextTaskFunction, rqptr);
      return;
   }

   /**************/
   /* statistics */
   /**************/

   vecptr = FaoVector;

   /* connections */
   *vecptr++ = Accounting.ConnectCount;
   *vecptr++ = Accounting.ConnectAcceptedCount;
   *vecptr++ = Accounting.ConnectRejectedCount;
   *vecptr++ = Accounting.ConnectCurrent;
   *vecptr++ = Accounting.ConnectPeak;
   *vecptr++ = Accounting.ConnectTooBusyCount;
   *vecptr++ = Accounting.ConnectSslCount;

   /* requests */
   *vecptr++ = Accounting.RequestTotalCount;
   *vecptr++ = Accounting.RequestParseCount;
   *vecptr++ = Accounting.RequestErrorCount;
   *vecptr++ = Accounting.RequestForbiddenCount;
   *vecptr++ = Accounting.RedirectRemoteCount;
   *vecptr++ = Accounting.RedirectLocalCount;
   *vecptr++ = Accounting.RequestKeepAliveCount;
   *vecptr++ = Accounting.RequestKeepAliveMax;
   *vecptr++ = Accounting.RequestPragmaNoCacheCount;

   CommaNumber (64, &Accounting.QuadBytesRawRx, BytesRawRx);
   CommaNumber (64, &Accounting.QuadBytesRawTx, BytesRawTx);

   *vecptr++ = BytesRawRx;
   *vecptr++ = BytesRawTx;

   /* HTTP methods */
   *vecptr++ = Accounting.MethodGetCount;
   *vecptr++ = Accounting.MethodHeadCount;
   *vecptr++ = Accounting.MethodPostCount;
   *vecptr++ = Accounting.MethodPutCount;

   /* responses */
   *vecptr++ = Accounting.ResponseStatusCodeCountArray[1];
   *vecptr++ = Accounting.ResponseStatusCodeCountArray[2];
   *vecptr++ = Accounting.ResponseStatusCodeCountArray[3];
   *vecptr++ = Accounting.ResponseStatusCodeCountArray[4];
   *vecptr++ = Accounting.ResponseStatusCodeCountArray[5];
   *vecptr++ = Accounting.ResponseStatusCodeCountArray[0];

   /* response duration */
   if (Accounting.ResponseDurationMin == -1)
   {
      *vecptr++ = 0;
      *vecptr++ = DURATION_DECIMAL_PLACES;
      *vecptr++ = 0;
   }
   else
   {
      *vecptr++ = Accounting.ResponseDurationMin / USECS_IN_A_SECOND;
      *vecptr++ = DURATION_DECIMAL_PLACES;
      *vecptr++ = (Accounting.ResponseDurationMin % USECS_IN_A_SECOND) /
                  DURATION_UNITS_USECS;
   }
   *vecptr++ = Accounting.ResponseDurationMax / USECS_IN_A_SECOND;
   *vecptr++ = DURATION_DECIMAL_PLACES;
   *vecptr++ = (Accounting.ResponseDurationMax % USECS_IN_A_SECOND) /
               DURATION_UNITS_USECS;
   if (Accounting.ResponseDurationCount)
   {
      status = lib$ediv (&Accounting.ResponseDurationCount,
                         &Accounting.QuadResponseDuration,
                         &ResponseDuration, &Remainder);
      if (Debug) fprintf (stdout, "lib$ediv() %%X%08.08X\n", status);
   }
   else
      ResponseDuration = 0;
   *vecptr++ = ResponseDuration / USECS_IN_A_SECOND;
   *vecptr++ = DURATION_DECIMAL_PLACES;
   *vecptr++ = (ResponseDuration % USECS_IN_A_SECOND) / DURATION_UNITS_USECS;

   /* modules */
   *vecptr++ = Accounting.DoServerAdminCount;
   *vecptr++ = Accounting.DoDirectoryCount;
   *vecptr++ = Accounting.DoFileCount;
   *vecptr++ = Accounting.DoFileNotModifiedCount;
   *vecptr++ = Accounting.DoIsMapCount;
   *vecptr++ = Accounting.DoMenuCount;
   *vecptr++ = Accounting.DoPutCount;
   *vecptr++ = ProxyAccounting.MethodConnectCount +
               ProxyAccounting.MethodGetCount +
               ProxyAccounting.MethodHeadCount +
               ProxyAccounting.MethodPostCount +
               ProxyAccounting.MethodPutCount,
   *vecptr++ = Accounting.DoSsiCount;
   *vecptr++ = Accounting.DoUpdateCount;
   *vecptr++ = Accounting.DoScriptCount;
   *vecptr++ = Accounting.DoCgiPlusScriptCount;
   *vecptr++ = Accounting.DclCgiPlusReusedCount;
   *vecptr++ = Accounting.DoAutoScriptCount;
   *vecptr++ = Accounting.DoDclCommandCount;
   *vecptr++ = Accounting.DclSpawnCount;
   *vecptr++ = Accounting.DoDECnetCount;
   *vecptr++ = Accounting.DoDECnetCount -
               Accounting.DoDECnetCgiCount -
               Accounting.DoDECnetOsuCount;
   *vecptr++ = Accounting.DoDECnetCgiCount;
   *vecptr++ = Accounting.DoDECnetOsuCount;

   /* cache */
   *vecptr++ = CacheMemoryInUse >> 10;
   *vecptr++ = CacheTotalKBytesMax;
   *vecptr++ = CacheLoadCount;
   *vecptr++ = CacheHits0;
   if (CacheLoadCount)
      *vecptr++ = CacheHits0 * 100 / CacheLoadCount;
   else
      *vecptr++ = 0;
   *vecptr++ = CacheHitCount;
   if (CacheLoadCount)
      *vecptr++ = CacheHitCount * 100 / CacheLoadCount;
   else
      *vecptr++ = 0;
   *vecptr++ = CacheHits10;
   if (CacheHitCount)
      *vecptr++ = CacheHits10 * 100 / CacheHitCount;
   else
      *vecptr++ = 0;
   *vecptr++ = CacheHits100;
   if (CacheHitCount)
      *vecptr++ = CacheHits100 * 100 / CacheHitCount;
   else
      *vecptr++ = 0;
   *vecptr++ = CacheHits100plus;
   if (CacheHitCount)
      *vecptr++ = CacheHits100plus * 100 / CacheHitCount;
   else
      *vecptr++ = 0;

   /* authorization */
   /*
      Note that the total number of paths being authorized may be
      greater than the total number of authentications.  This happens
      if a request contains both script and path authorizations,
      resulting in two authorizations for the one request!
   */
   *vecptr++ = Accounting.AuthAuthorizedCount;
   *vecptr++ = Accounting.AuthNotAuthorizedCount;
   *vecptr++ = Accounting.AuthBasicCount +
               Accounting.AuthDigestCount -
               Accounting.AuthAgentCount -
               Accounting.AuthHtDatabaseCount -
               Accounting.AuthSimpleListCount -
               Accounting.AuthVmsCount;
   *vecptr++ = Accounting.AuthAgentCount;
   *vecptr++ = Accounting.AuthHtDatabaseCount;
   *vecptr++ = Accounting.AuthSimpleListCount;
   *vecptr++ = Accounting.AuthVmsCount;
   *vecptr++ = Accounting.AuthBasicCount;
   *vecptr++ = Accounting.AuthDigestCount;

   status = sys$faol (&StatisticsFaoDsc, &Length, &BufferDsc, &FaoVector);
   if (VMSnok (status) || status == SS$_BUFFEROVF)
   {
      rqptr->ErrorTextPtr = "sys$faol()";
      ErrorVmsStatus (rqptr, status, FI_LI);
      SysDclAst (NextTaskFunction, rqptr);
      return;
   }

   Buffer[Length] = '\0';
   NetWriteBuffered (rqptr, 0, Buffer, Length);

   /* proxy serving */
   status = ProxyMaintStatisticsReport (rqptr);
   if (VMSnok (status) || status == SS$_BUFFEROVF)
   {
      rqptr->ErrorTextPtr = "sys$faol()";
      ErrorVmsStatus (rqptr, status, FI_LI);
      SysDclAst (NextTaskFunction, rqptr);
      return;
   }

   /* other */
   vecptr = FaoVector;
   *vecptr++ = Accounting.StreamLfConversionCount;

   status = sys$faol (&OtherFaoDsc, &Length, &BufferDsc, &FaoVector);
   if (VMSnok (status) || status == SS$_BUFFEROVF)
   {
      rqptr->ErrorTextPtr = "sys$faol()";
      ErrorVmsStatus (rqptr, status, FI_LI);
      SysDclAst (NextTaskFunction, rqptr);
      return;
   }

   Buffer[Length] = '\0';
   NetWriteBuffered (rqptr, NextTaskFunction, Buffer, Length);
}

/*****************************************************************************/
/*
Execute server administration menu control actions.
*/ 

AdminControl
(
struct RequestStruct *rqptr,
void *NextTaskFunction
)
{
   int  status;
   char  String [256];

   /*********/
   /* begin */
   /*********/

   if (Debug) fprintf (stdout, "AdminControl()\n");

   if (strsame (rqptr->PathInfoPtr, INTERNAL_CONTROL_DECNET_PURGE, -1))
   {
      /********************************/
      /* purge network scripting task */
      /********************************/

      DECnetDisconnect (rqptr, NextTaskFunction, false);
      return;
   }
   else
   if (strsame (rqptr->PathInfoPtr, INTERNAL_CONTROL_DECNET_FORCE, -1))
   {
      /***********************************************/
      /* forcably disconnect network scripting tasks */
      /***********************************************/

      DECnetDisconnect (rqptr, NextTaskFunction, true);
      return;
   }
   else
   if (strsame (rqptr->PathInfoPtr, INTERNAL_CONTROL_EXIT, -1))
   {
      /***************/
      /* server exit */
      /***************/

      if (ConnectCountCurrent > 1)
      {         
         sprintf (String,
           "exit when %d other, in-progress request(s) complete.</FONT>",
           ConnectCountCurrent-1);
         AdminControlSuccess (rqptr, "<FONT COLOR=\"#ff0000\">", String);

         /* stop the server from receiving incoming requests */
         NetShutdownServerSocket ();
         ControlExitRequested = true;
      }
      else
      {
         AdminControlSuccess (rqptr, "<FONT COLOR=\"#ff0000\">",
                              "exit now!</FONT>");
         fprintf (stdout, "%%%s-I-ADMIN, %s, server exit\n",
                  Utility, DateTime(NULL,20));
         ExitStatus = SS$_NORMAL;
         HttpdExit (&ExitStatus);
         sys$delprc (0, 0);
      }
   }
   else
   if (strsame (rqptr->PathInfoPtr, INTERNAL_CONTROL_EXIT_NOW, -1))
   {
      /********************/
      /* server exit NOW! */
      /********************/

      AdminControlSuccess (rqptr, "<FONT COLOR=\"#ff0000\">",
                                  "exit NOW!</FONT>");

      fprintf (stdout, "%%%s-I-ADMIN, %s, server exit NOW!\n",
               Utility, DateTime(NULL,20));
      ExitStatus = SS$_NORMAL;
      HttpdExit (&ExitStatus);
      sys$delprc (0, 0);
   }
   else
   if (strsame (rqptr->PathInfoPtr, INTERNAL_CONTROL_RESTART, -1))
   {
      /******************/
      /* server restart */
      /******************/

      if (ConnectCountCurrent > 1)
      {
         sprintf (String,
            "restart when %d other, in-progress request(s) complete.</FONT>",
            ConnectCountCurrent-1);
         AdminControlSuccess (rqptr, "<FONT COLOR=\"#ff0000\">", String);

         /* stop the server from receiving incoming requests */
         NetShutdownServerSocket ();
         ControlRestartRequested = true;
      }
      else
      {
         AdminControlSuccess (rqptr, "<FONT COLOR=\"#ff0000\">",
                                     "restart now!</FONT>");
         fprintf (stdout, "%%%s-I-ADMIN, %s, server restart\n",
                  Utility, DateTime(NULL,20));
         exit (SS$_NORMAL);
      }
   }
   else
   if (strsame (rqptr->PathInfoPtr, INTERNAL_CONTROL_RESTART_NOW, -1))
   {
      /***********************/
      /* server restart NOW! */
      /***********************/

      AdminControlSuccess (rqptr, "<FONT COLOR=\"#ff0000\">",
                                  "restart NOW!</FONT>");
      fprintf (stdout, "%%%s-I-ADMIN, %s, server restart NOW!\n",
               Utility, DateTime(NULL,20));
      exit (SS$_NORMAL);
   }
   else
   if (strsame (rqptr->PathInfoPtr, INTERNAL_CONTROL_DCL_PURGE, -1))
   {
      /**************************/
      /* purge DCL subprocesses */
      /**************************/

      DclPurgeAllSubprocesses (rqptr, NextTaskFunction, false);
      return;
   }
   else
   if (strsame (rqptr->PathInfoPtr, INTERNAL_CONTROL_DCL_DELETE, -1))
   {
      /************************************/
      /* forcably delete DCL subprocesses */
      /************************************/

      DclPurgeAllSubprocesses (rqptr, NextTaskFunction, true);
      return;
   }
   else
   if (strsame (rqptr->PathInfoPtr, INTERNAL_CONTROL_MAPPING_LOAD, -1))
   {
      /************************/
      /* reload mapping rules */
      /************************/

      fprintf (stdout, "%%%s-I-ADMIN, %s, mapping rules reload\n",
               Utility, DateTime(NULL,20));

      MapUrl_ControlReload (rqptr, NextTaskFunction);
      return;
   }
   else
   if (strsame (rqptr->PathInfoPtr, INTERNAL_CONTROL_AUTH_LOAD, -1))
   {
      /*****************************/
      /* authorization path reload */
      /*****************************/

      fprintf (stdout, "%%%s-I-ADMIN, %s, authorization paths reload\n",
               Utility, DateTime(NULL,20));

      AuthConfigReload (rqptr, NextTaskFunction);
      return;
   }
   else
   if (strsame (rqptr->PathInfoPtr, INTERNAL_CONTROL_AUTH_PURGE, -1))
   {
      /******************************/
      /* purge authentication cache */
      /******************************/

      fprintf (stdout, "%%%s-I-ADMIN, %s, purge authentication cache\n",
               Utility, DateTime(NULL,20));

      /* fake it :^) */
      rqptr->QueryStringPtr = "do=purge";
      HTAdminBegin (rqptr, NextTaskFunction);
      return;
   }
   else
   if (strsame (rqptr->PathInfoPtr, INTERNAL_CONTROL_ZERO, -1))
   {
      /*******************/
      /* zero accounting */
      /*******************/

      fprintf (stdout, "%%%s-I-ADMIN, %s, accounting zeroed\n",
               Utility, DateTime(NULL,20));

      ZeroAccounting ();
      AdminControlSuccess (rqptr, "", "accounting zeroed.");
   }
   else
   if (strsame (rqptr->PathInfoPtr, INTERNAL_CONTROL_CACHE_ON, -1))
   {
      /****************/
      /* enable cache */
      /****************/

      if (CacheEnabled)
         ErrorGeneral (rqptr, "Cache already enabled.", FI_LI);
      else
      {
         CacheEnabled = true;
         AdminControlSuccess (rqptr, "", "cache enabled.");
      }
   }
   else
   if (strsame (rqptr->PathInfoPtr, INTERNAL_CONTROL_CACHE_OFF, -1))
   {
      /*****************/
      /* disable cache */
      /*****************/

      if (CacheEnabled)
      {
         CacheEnabled = false;
         AdminControlSuccess (rqptr, "", "cache disabled.");
      }
      else
         ErrorGeneral (rqptr, "Cache already disabled", FI_LI);
   }
   else
   if (strsame (rqptr->PathInfoPtr, INTERNAL_CONTROL_CACHE_PURGE, -1))
   {
      /***************/
      /* purge cache */
      /***************/

      if (CacheEnabled)
      {
         int  PurgeCount, MarkedForPurgeCount;

         CachePurge (false, &PurgeCount, &MarkedForPurgeCount);
         sprintf (String, "cache %d purged, %d marked for purge.",
                  PurgeCount, MarkedForPurgeCount);
         AdminControlSuccess (rqptr, "", String);
      }
      else
         ErrorGeneral (rqptr, "Cache not enabled.", FI_LI);
   }
   else
   if (strsame (rqptr->PathInfoPtr, INTERNAL_CONTROL_LOG_ON, -1))
   {
      /******************/
      /* enable logging */
      /******************/

      if (LoggingEnabled)
         ErrorGeneral (rqptr, "Logging already enabled.", FI_LI);
      else
      if (VMSok (status = Logging (NULL, LOGGING_OPEN)))
      {
         fprintf (stdout, "%%%s-I-ADMIN, %s, logging enabled\n",
                  Utility, DateTime(NULL,20));
         AdminControlSuccess (rqptr, "", "logging enabled.");
      }
      else
      {
         rqptr->ErrorTextPtr = "Logging could NOT be enabled";
         ErrorVmsStatus (rqptr, status, FI_LI);
      }
   }
   else
   if (strsame (rqptr->PathInfoPtr, INTERNAL_CONTROL_LOG_OFF, -1))
   {
      /*******************/
      /* disable logging */
      /*******************/

      if (!LoggingEnabled)
         ErrorGeneral (rqptr, "Logging not currently enabled.", FI_LI);
      else
      if (VMSok (status = Logging (NULL, LOGGING_CLOSE)))
      {
         fprintf (stdout, "%%%s-I-ADMIN, %s, logging disabled\n",
                  Utility, DateTime(NULL,20));
         AdminControlSuccess (rqptr, "", "logging disabled.");
      }
      else
      {
         rqptr->ErrorTextPtr = "Logging could NOT be disabled!";
         ErrorVmsStatus (rqptr, status, FI_LI);
      }
   }
   else
   if (strsame (rqptr->PathInfoPtr, INTERNAL_CONTROL_LOG_FLUSH, -1))
   {
      /*************/
      /* flush log */
      /*************/

      if (!LoggingEnabled)
         ErrorGeneral (rqptr, ErrorLoggingNotEnabled, FI_LI);
      else
      if (VMSok (status = Logging (NULL, LOGGING_FLUSH)))
      {
         fprintf (stdout, "%%%s-I-ADMIN, %s, log flushed\n",
                  Utility, DateTime(NULL,20));
         AdminControlSuccess (rqptr, "", "log flushed.");
      }
      else
      {
         rqptr->ErrorTextPtr = "Log could NOT be flushed!";
         ErrorVmsStatus (rqptr, status, FI_LI);
      }
   }
   else
   if (strsame (rqptr->PathInfoPtr, INTERNAL_CONTROL_PROXY, -1))
   {
      /*****************/
      /* proxy serving */
      /*****************/

      ProxyMaintControl (rqptr);
   }
   else
   if (strsame (rqptr->PathInfoPtr, INTERNAL_CONTROL_PROXY_ADJUST, -1))
   {
      /**************************************/
      /* adjust proxy serving configuration */
      /**************************************/

      ProxyMaintAdjust (rqptr, NextTaskFunction);
      return;
   }
   else
   {
      /********************/
      /* unknown function */
      /********************/

      rqptr->ResponseStatusCode = 403;
      ErrorGeneral (rqptr, "Unknown control function.", FI_LI);
   }

   SysDclAst (NextTaskFunction, rqptr);
}

/*****************************************************************************/
/*
Return a 200 success response to the client.
*/ 

int AdminControlSuccess
(
struct RequestStruct *rqptr,
char *StartText,
char *EndText
)
{
   static $DESCRIPTOR (ResponseFaoDsc,
"<HTML>\n\
<HEAD>\n\
!AZ\
<TITLE>Success 200</TITLE>\n\
</HEAD>\n\
<BODY>\n\
<H1>SUCCESS!!</H1>\n\
<P>Reported by server.\n\
<P>!AZServer !AZ !AZ\n\
</BODY>\n\
</HTML>\n");

   register unsigned long  *vecptr;

   int  status;
   unsigned long  FaoVector [16];
   unsigned short  Length;
   char  Buffer [2048];
   $DESCRIPTOR (BufferDsc, Buffer);

   /*********/
   /* begin */
   /*********/

   if (Debug) fprintf (stdout, "AdminControlSuccess()\n");

   rqptr->ResponsePreExpired = PRE_EXPIRE_ADMIN;
   rqptr->ResponseStatusCode = 200;
   HTTP_HEADER_200_HTML (rqptr);

   vecptr = FaoVector;

   *vecptr++ = HtmlMetaInfo (rqptr, NULL);
   *vecptr++ = StartText;
   *vecptr++ = ServerHostPort;
   *vecptr++ = EndText;
   sys$faol (&ResponseFaoDsc, &Length, &BufferDsc, &FaoVector);
   Buffer[Length] = '\0';

   NetWrite (rqptr, 0, Buffer, Length);

   return (SS$_NORMAL);
}

/****************************************************************************/

