/*****************************************************************************/
/*
                                  HTTPd.c

WASD VMS Hypertext Transfer Protocol daemon.


COPYRIGHT NOTICE
----------------
WASD VMS Hypertext Services, Copyright (C) 1996-1999 Mark G.Daniel.
(prior to 01-JUL-1997, "HFRD VMS Hypertext Services")

This package is free software; you can redistribute it and/or modify it under 
the terms of the GNU General Public License as published by the Free Software 
Foundation; version 2 of the License, or any later version. 

This package is distributed in the hope that it will be useful, but WITHOUT 
ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS 
FOR A PARTICULAR PURPOSE.  See the GNU General Public License for more 
details. 

You should have received a copy of the GNU General Public License along with 
this package; if not, write to the Free Software Foundation, Inc., 675 Mass 
Ave, Cambridge, MA 02139, USA. 


PRIVILEGES REQUIRED
-------------------
These should be INSTALLed against the image.

ALTPRI   Allows the server account to raise it's prioity above 4 if enabled by
         the /PRIORITY= qualifier.

SYSPRV   Used for various purposes, including creating sockets within the
         privileged port range (1-1023, which includes port 80 of course),
         accessing configuration files (which can be protected from world
         access), enable AUTHORIZED write access to the file system, checking
         authentication details, amongst others.

SYSNAM   Allows the server to write into the LNM$SYSTEM logical name table.

PRMMBX   Used by the subprocess scripting module to create permanent mailboxes
         (much more efficient than creating a new set with each subprocess).

PSWAPM   Allows the server process to prevent itself from being swapped out if 
         enabled by the /[NO]SWAP qualifier.


QUALIFIERS  (implemented in the CLI.c module)
----------
/ACCEPT=                comma separated list of accepted hosts/domains
/AUTHORIZE=[SSL|ALL]    authorization may only be used with "https:",
                        all paths must be authorized
/CGI_PREFIX=            prefix to CGI-script variable (symbol) names
/DBUG                   turn on all "if (Debug)" statements (if compiled DBUG)
/DO=                    command to be passed to server
/FILBUF=                number of bytes in record buffer
/FORMAT=                log record format
/REJECT=                comma separated list of rejected hosts/domains
/[NO]LOG[=]             logging enabled/disabled, optional log file name
/[NO]MONITOR            monitor logicals enabled/disabled
/NETBUF=                number of bytes in network read buffer
/OUTBUF=                number of bytes in output buffer
/PERIOD=                log file period
/PORT=                  server IP port number (overriden by [service] config)
/PRIORITY=              process priority
/[NO]PROFILE            allow/disallow SYSUAF-authenticated access control
/[NO]PROMISCUOUS[=pwd]  authenticates any user name for path authorization
                        (optional password required for authorization) 
/SERVICES=              list of [host-name:]port providing HTTP service
/SOFTWAREID=            overrides the server's software ID string
/SYSPRV                 normally operate with SYSPRV enabled (CAUTION!)
/[NO]SSL=               enables Secure Sockets Layer, sets protocol parameters
/[NO]SYSUAF[=ID,SSL,RELAXED]    allow/disallow SYSUAF authentication, or
                        SYSUAF authentication is allowed by identifier,
                        SYSUAF authentication is only allowed with "https:",
                        SYSUAF authentication allowed with any current account
/[NO]SWAP               allow/disallow process to be swapped out
/VERSION                simply display the server version
/[NO]ZERO               accounting zeroed on startup (default is non-zeroed)


VERSION HISTORY
---------------
06-JAN-2000  MGD  v6.1.2,
                  authorization failure limit evasion period,
                  numerous warnings from DECC v6.2 addressed,
                  bugfix; user restriction list pass (broken in 6.1)
17-DEC-1999  MGD  v6.1.1,
                  bugfix; quote double-up in CgiVariable() (INSVIRMEM exit)
04-DEC-1999  MGD  v6.1.0,
                  "agent" authentication/authorization,
                  CGI(plus) processing provides callouts,
                  SSI module now supports OSU-specific directives,
                  /SYSPRV now allows operation with SYSPRV turned on,
                  "one-shot" WATCH and "peek" reports,
                  output no-progress timer,
                  remove NETLIB support
16-OCT-1999  MGD  v6.0.3,
                  bugfix; sys$create_user_profile,
                  bugfix; mapping storage overflow
                  USER mapping rule for SYSUAF access
12-SEP-1999  MGD  v6.0.2,
                  minor changes to authorization processing,
                  bugfix; service parsing and SSL,
                  virtual services now match using "Host:" field
19-JUN-1999  MGD  v6.0.1,
                  refinements to request termination/rundown,
                  bugfix; DECnet (CGI and OSU) task handling
                  bugfix; proxy request HTTP/0.9 response processing
30-MAY-1999  MGD  v6.0.0,
                  proxy, with HTTP caching,
                  OpenSSL 0.9.3 support (also SSLeay support),
                  extended authorization/authentication environment
31-MAR-1999  MGD  v5.3.4,
                  bugfix; SesolaReport(), HttpHeaderChallenge()
28-MAR-1999  MGD  v5.3.3,
                  SSI variables global (when "#include"ing other SSI),
                  SSI read buffer determined by 'FileXabFhc.xab$w_lrl'
05-FEB-1999  MGD  v5.3.2,
                  bugfix; FileNextRecord() zero '_usz'
10-JAN-1999  MGD  v5.3.1,
                  greater granularity when WATCHing authorization,
                  bugfix; OSU scripting pass *mapped* file spec
14-NOV-1998  MGD  v5.3.0,
                  [[host:port]] virtual service syntax,
                  [AddType] can now "text/html; charset=ISO-8859-1",
                  [CharsetDefault] sets text and server character set,
                  improved AST granularity several significant modules,
                  WATCH report and CLI,
                  RMS-invalid substitution character in mapping rules,
                  bugfix; NameOfDirectoryFile()
29-AUG-1998  MGD  v5.2.0,
                  reuse DECnet task connections,
                  allow specified hosts exclusion from logging,
                  stream-LF conversion only on specified paths,
                  bugfix; SYS$TIMEZONE_DIFFERENTIAL processing
                  bugfix; DECnet tasks not aborted at timeout
07-JUL-1998  MGD  v5.1.0,
                  add eXtended Server Side Includes processing,
                  design-problem; modify CGIplus script rundown,
                  SYSUAF authentication by identifier,
                  per-service logging,
                  rqptr->TimerTerminated (occasional lib$get_vm()
                  %LIB-F-BADLOADR around connection expiry termination)
20-DEC-1997  MGD  v5.0.0,
                  optional Secure Sockets Layer (using SSLeay),
                  DECnet-based scripting including OSU emulation,
                  miscellaneous revisions and "improvements"
07-JAN-1997  MGD  v4.5.2,
                  bugfix; record-mode file transfer,
                  bugfix; activity graph
06-DEC-1997  MGD  v4.5.1,
                  resolving a suspected inconsistent AST delivery situation
                  by requiring all $QIO()s with AST routines to ensure any
                  queueing errors, etc. are reported via the AST routine, by
                  an explicit $DCLAST() ... this removes ambiguity about how
                  $QIO() returns should be handled ... drastic, but desperate
                  times, etc. (a more consistent and desirable model anyway :^)
02-NOV-1997  MGD  v4.5.0,
                  file cache,
                  logging periods,
                  HttpdSupervisor(),
                  configurable script run-time environments,
                  additional request header fields
18-OCT-1997  MGD  v4.4.1,
                  bugfix; duration,
                  bugfix; logging period
01-OCT-1997  MGD  v4.4.0,
                  message module,
                  conditional rule mapping,
                  SYSUAF-authenticated user access control,
                  multi-homed/multi-port services
                  (some NETLIB packages now cannot DNS lookup),
                  echo and Xray internal scripts,
                  extensions to logging functionality,
                  additional command-line server control,
                  bugfix; redirection loop detection
01-AUG-1997  MGD  v4.3.0,
                  MadGoat NETLIB broadens TCP/IP package support,
                  server activity report
16-JUL-1997  MGD  v4.2.2,
                  bugfix; WORLD realm and access list
07-JUL-1997  MGD  v4.2.1,
                  minimum heap allocation chunk size,
                  prevent keep-alive timeout redefining request logical
01-JUL-1997  MGD  v4.2.0,
                  change name to WASD (Wide Area Surveillance Division)
                  persistant DCL subprocesses and CGIplus 
                  (see re-written DCL.C module),
                  scripting and client reports,
                  potential multi-thread problems in reports fixed
27-MAR-1997  MGD  v4.1.0,
                  rationalized HTTP response header generation,
                  delete on close for "temporary" files, to support
                  UPD module "preview" functionality ... WARNING, any
                  file with a name comprising a leading hyphen,
                  sixteen digits and a trailing hyphen will be deleted!
01-FEB-1997  MGD  v4.0.0,
                  HTTPd version 4
01-OCT-1996  MGD  v3.4.0,
                  extended server reporting  
01-AUG-1996  MGD  v3.3.0,
                  realm/path-based authorization,
                  BASIC and DIGEST authentication,
                  PUT(/POST/DELETE) module,
                  StmLf module (variable to stream-LF file conversion)
12-APR-1996  MGD  v3.2.0,
                  file record/binary now determined by record format,
                  persistent connections ("Keep-Alive" within HTTP/1.0),
                  moved RMS parse structures into thread data,
                  improved local redirection detection,
                  observed Multinet disconnection/zero-byte behaviour
                  (request now aborts if network read returns zero bytes)
15-FEB-1996  MGD  v3.1.1,
                  fixed rediculous :^( bug in 302 HTTP header,
                  minor changes to request accounting and server report,
                  minor changes for user directory support,
                  minor changes to error reporting
03-JAN-1996  MGD  v3.1.0,
                  support for both DEC TCP/IP Services and TGV MultiNet
01-DEC-1995  MGD  v3.0.0,
                  single heap for each thread's dynamic memory management,
                  extensive rework of DCL subprocess functionality,
                  HTML pre-processsing module (aka Server Side Includes),
                  NCSA/CERN compliant image-mapping module,
                  NetWriteBuffered() for improving network IO,
                  miscellaneous reworks/rewrites
27-SEP-1995  MGD  v2.3.0,
                  carriage-control on non-header records from <CR><LF> 
                  to single <LF> ('\n' ... newline), some browsers expect
                  only this (e.g. Netscape 1.n was spitting on X-bitmaps)
                  added Greenwich Mean Time time-stamp functionality,
                  added 'Referer:', 'If-Modified-Since:', 'User-Agent:'
07-AUG-1995  MGD  v2.2.2,
                  optionally include commented VMS file specifications
                  in HTML documents and VMS-style directory listings
16-JUN-1995  MGD  v2.2.1,
                  added file type description to "Index of" (directory)
24-MAY-1995  MGD  v2.2.0,
                  minor changes to allow compilation on AXP platform
03-APR-1995  MGD  v2.1.0,
                  add SYSUAF authentication, POST method handling
20-DEC-1994  MGD  v2.0.0,
                  multi-threaded version
20-JUN-1994  MGD  v1.0.0,
                  single-threaded version
*/
/*****************************************************************************/

/* standard C header files */
#include <ctype.h>
#include <errno.h>
#include <stdio.h>
#include <string.h>

/* VMS related header files */
#include <descrip.h>
#include <jpidef.h>
#include <libdef.h>
#include <prvdef.h>
#include <psldef.h>
#include <ssdef.h>
#include <stsdef.h>
#include <syidef.h>

/* application-related header files */
#include "wasd.h"
#include "copyright.h"

#define WASD_MODULE "HTTPD"

/******************/
/* global storage */
/******************/

/*
   These are the required privileges of the executing HTTP server.
   The server ACCOUNT should only have TMPMBX and NETMBX (just for
   extra security, policy ... keep privileged accounts to a minimum).
   The image should be installed with SYSPRV, PRMMBX, PSWAPM and SYSNAM.
   Subprocesses are spawned only with the process's current privileges,
   which are always maintained at TMPMBX and NETMBX.  If additional
   privileges are required for any particular purpose (e.g. binding to
   a privileged IP port) then they are enabled, the action performed,
   and then they are disabled again immediately.
*/
unsigned long  AltPriMask [2] = { PRV$M_ALTPRI, 0 },
               AveJoePrvMask [2] = { PRV$M_NETMBX | PRV$M_TMPMBX, 0 },
               PswapmMask [2] = { PRV$M_PSWAPM, 0 },
               ResetPrvMask [2] = { 0xffffffff, 0xffffffff },
               SysPrvMask [2] = { PRV$M_SYSPRV, 0 };

char  Utility [] = "HTTPD";

/*
   Decided to be able to eliminate the debug statements from production
   executables completely.  This will eliminate some largely unused code
   from the images reducing the overall file size, but more importantly
   will eliminate the test and branch execution overhead each time a debug
   statement is encountered.  Do this by conditionally turning the integer
   Debug storage into a constant false value ... compiler optimization of
   an impossible-to-execute section of code does the rest!  Very little
   other code needs to be explicitly conditionally compiled.
*/
#ifdef DBUG
boolean  Debug;
#else
#define Debug 0
#endif

boolean  MonitorEnabled,
         NaturalLanguageEnglish,
         NoSwapOut = true,
         OperateWithSysPrv = false;

int  ExitStatus,
     FileBufferSize = DEFAULT_FILE_BUFFER_SIZE,
     ProcessPriority = 4,
     ServerPort = 80;

unsigned long  HttpdProcessId;

char  CommandLine [256],
      HttpdUserName [13],
      NaturalLanguage [64],
      ServerPortString [16],
      SyiHwName [32],
      SyiVersion [9];

struct AnExitHandler  ExitHandler;

struct AccountingStruct  Accounting;
boolean  AccountingZeroOnStartup;

#ifdef DBUG
   USEC_TIMER_STORAGE
#endif

/********************/
/* external storage */
/********************/

extern boolean  AuthorizationEnabled,
                CliMonitorDisabled,
                CliMonitorEnabled,
                CliVersion,
                ControlExitRequested,
                ControlRestartRequested,
                LoggingEnabled,
                ProtocolHttpsConfigured;

extern int  CliServerPort,
            ConnectCountConcurrentMax,
            HttpdAccountingVersion,
            RequestHistoryMax,
            ServerPort,
            WatchEnabled;

extern char  BuildInfo[],
             CliLogFileName[],
             ControlBuffer[],
             ErrorSanityCheck[],
             HttpdIpPackage[],
             HttpdSesola[],
             SoftwareID[],
             TimeGmtString[];

extern struct ConfigStruct  Config;
extern struct MsgDbStruct  MsgDb;
extern struct ListHeadStruct  RequestList;

/* for initial debugging only */
extern unsigned long  VmCacheVmZoneId,
                      VmGeneralVmZoneId,
                      VmRequestVmZoneId;

/*****************************************************************************/
/*
*/

int main ()

{
   static unsigned short  SyiHwNameLength;
   static unsigned long  JpiGrp,
                         SyiMaxSysGroup;

   static struct {
      unsigned short  BufferLength;
      unsigned short  ItemCode;
      unsigned long  BufferAddress;
      unsigned long  ReturnLengthAddress;
   }
      JpiItem [] =
   { { sizeof(JpiGrp), JPI$_GRP, &JpiGrp, 0 },
     { sizeof(HttpdProcessId), JPI$_PID, &HttpdProcessId, 0 },
     { sizeof(HttpdUserName)-1, JPI$_USERNAME, &HttpdUserName, 0 },
     { 0,0,0,0 }
   },
      SyiItem [] =
   { { sizeof(SyiMaxSysGroup), SYI$_MAXSYSGROUP, &SyiMaxSysGroup, 0 },
     { sizeof(SyiVersion)-1, SYI$_VERSION, &SyiVersion, 0 },
     { sizeof(SyiHwName)-1, SYI$_HW_NAME, &SyiHwName, &SyiHwNameLength },
     { 0,0,0,0 }
   };

   register char  *cptr, *sptr, *zptr;

   int  status;
   unsigned short  Length;
   char  TimeScratch [32],
         ProcessName [16];
   $DESCRIPTOR (NaturalLanguageDsc, NaturalLanguage);
   $DESCRIPTOR (ProcessNameDsc, ProcessName);
   $DESCRIPTOR (ProcessNameFaoDsc, "HTTPd:!UL");

   /*********/
   /* begin */
   /*********/

   if (VMSnok (status = ParseCommandLine ()))
      exit (status);

    VersionInfo ();

#ifdef DBUG
   if (!Debug)
      if ((char*)getenv ("HTTPD$DBUG") != NULL)
         Debug = true;
#endif

   if (ControlBuffer[0])
   {
      /*********************************************************/
      /* this image is being used to control the HTTPd process */
      /*********************************************************/

      if (CliServerPort) ServerPort = CliServerPort;
      if (ServerPort < 1 || ServerPort > 65535)
         ErrorExitVmsStatus (0, "IP port", FI_LI);
      exit (ControlCommand (ControlBuffer));
   }

   /****************/
   /* HTTPd server */
   /****************/

   /* output the GNU GENERAL PUBLIC LICENSE message */
   fprintf (stdout, "%%%s-I-SOFTWAREID, %s\n%s",
            Utility, SoftwareID, CopyRightMessageBrief);

   if (CliVersion) exit (SS$_NORMAL);

   /* timestamp the server startup */
   fprintf (stdout, "%%%s-I-STARTUP, %s\n", Utility, DateTime(NULL,20));

   /* display information about any CLI WATCH facility settings */
   WatchCliSettings ();

   /* initialize general-use virtual memory management */
   VmGeneralInit ();
   /** if (Debug) VmDebug (VmGeneralVmZoneId); **/

   /* get some account details the server is executing under */
   if (VMSnok (status = sys$getjpi (0, 0, 0, &JpiItem, 0, 0, 0)))
      ErrorExitVmsStatus (status, "sys$getjpi()", FI_LI);
   HttpdUserName[sizeof(HttpdUserName)-1] = '\0';
   for (cptr = HttpdUserName; *cptr && *cptr != ' '; cptr++);
   *cptr = '\0';
   if (Debug)
      fprintf (stdout, "UserName |%s| ProcessId: %08.08X Grp: %d\n",
              HttpdUserName, HttpdProcessId, JpiGrp);

   /* get system information */
   if (VMSnok (status = sys$getsyi (0, 0, 0, &SyiItem, 0, 0, 0)))
      ErrorExitVmsStatus (status, "sys$getsyi()", FI_LI);

   SyiHwName[SyiHwNameLength] = '\0';
   SyiVersion[sizeof(SyiVersion)-1] = '\0';
   for (cptr = SyiVersion; *cptr && *cptr != ' '; cptr++);
   *cptr = '\0';
   if (Debug)
      fprintf (stdout, "HwName |%s| SyiVersion |%s|\n",
               SyiHwName, SyiVersion);

   if (JpiGrp <= SyiMaxSysGroup)
      fprintf (stdout,
"%%%s-W-SYSPRV, operating with implicit SYSPRV (UIC group %d)\n",
               Utility, JpiGrp);

   /* make sure the process' privileges are those of a mere mortal */
   if (VMSnok (status = sys$setprv (0, &ResetPrvMask, 0, 0)))
      ErrorExitVmsStatus (status, "sys$setprv()", FI_LI);
   if (VMSnok (status = sys$setprv (1, &AveJoePrvMask, 0, 0)))
      ErrorExitVmsStatus (status, "sys$setprv()", FI_LI);

   if (OperateWithSysPrv && OPERATE_WITH_SYSPRV)
   {
      /* looks like we're been asked to use Superman's Xray vision! */
      if (VMSnok (status = sys$setprv (1, &SysPrvMask, 0, 0)))
         ErrorExitVmsStatus (status, "sys$setprv()", FI_LI);
      fprintf (stdout, "%%%s-W-SYSPRV, operating with explicit SYSPRV\n",
               Utility);
   }
   else
   if (OperateWithSysPrv)
   {
      OperateWithSysPrv = false;
      fprintf (stdout, "%%%s-W-SYSPRV, operation is NOT a compiled option\n",
               Utility);
   }

   /* set (a sensible) process priority */
   if (ProcessPriority > 6) ProcessPriority = 6;
   if (VMSnok (status = sys$setprv (1, &AltPriMask, 0, 0)))
      ErrorExitVmsStatus (status, "sys$setprv()", FI_LI);
   if (VMSnok (status = sys$setpri (0, 0, ProcessPriority, 0, 0, 0)))
      ErrorExitVmsStatus (status, "sys$setpri()", FI_LI);
   if (VMSnok (status = sys$setprv (0, &AltPriMask, 0, 0)))
      ErrorExitVmsStatus (status, "sys$setprv()", FI_LI);

   /* read the server configuration file */
   Configure (&Config);

   /* anywhere near before starting logging! */
   NetGetServerHostName ();

   /* initialize Secure Sockets Layer, if available and if required */
   SeSoLaInit ();

   if (CliServerPort)
      ServerPort = CliServerPort;
   else
   if (Config.ServerPort)
      ServerPort = Config.ServerPort;
   if (ServerPort < 1 || ServerPort > 65535)
      ErrorExitVmsStatus (0, "IP port", FI_LI);
   sprintf (ServerPortString, "%d", ServerPort);

   /* get and set the service names and ports */
   NetSetService ();

   /* initialize Secure Sockets Layer for each "https:" service (if any) */
   SeSoLaInitServiceList ();

   /* set the process name */
   if (VMSnok (status =
       sys$fao (&ProcessNameFaoDsc, &Length, &ProcessNameDsc, ServerPort)))
      ErrorExitVmsStatus (status, "sys$fao()", FI_LI);
   ProcessName[ProcessNameDsc.dsc$w_length = Length] = '\0';
   if (Debug) fprintf (stdout, "ProcessName |%s|\n", ProcessName);
   if (VMSnok (status = sys$setprn (&ProcessNameDsc)))
      ErrorExitVmsStatus (status, "sys$setprn()", FI_LI);

   /* ensure the GMT time/logical is available */
   if (VMSnok (status = TimeSetGMT ()))
      ErrorExitVmsStatus (status, "TimeSetGMT()", FI_LI);
   fprintf (stdout, "%%%s-I-GMT, %s\n", Utility, TimeGmtString);

   if (CliMonitorDisabled)
      MonitorEnabled = false;
   else
   if (Config.MonitorEnabled || CliMonitorEnabled)
      MonitorEnabled = true;
   else
      MonitorEnabled = false;

   /* reset logical names provided to the HTTPd monitor utility */
   if (MonitorEnabled)
      if (VMSnok (status = DefineMonitorLogicals (NULL)))
         ErrorExitVmsStatus (status, "DefineMonitorLogicals()", FI_LI);

   /* must be done after DefineMonitorLogicals(), which gets previous values */
   if (AccountingZeroOnStartup ||
       HttpdAccountingVersion != Accounting.AccountingVersion ||
       !MonitorEnabled)
   {
      /* initialize accounting information without reseting the logicals */
      ZeroAccounting ();
   }
   Accounting.AccountingVersion = HttpdAccountingVersion;

   /* initialize message database */
   MsgLoad (&MsgDb);

   /* authentication/authorization configuration */
   AuthConfigInit ();

   /* load the rule mapping database */
   MapUrl_Map (NULL, 0, NULL, 0, NULL, 0, NULL, 0, NULL);

   /* initialize proxy processing */
   ProxyInit ();

   /* initialize DCL processing */
   DclInit ();

   if (Config.Busy) ConnectCountConcurrentMax = Config.Busy;

   /* initialize request virtual memory management */
   VmRequestInit (ConnectCountConcurrentMax); 
   /** if (Debug) VmDebug (VmRequestVmZoneId); **/

   /* initialize file cache  */
   CacheInit (true);
   /** if (Debug) VmDebug (VmCacheVmZoneId); **/
   /** if (Debug) VmDebug (0); **/

   /* initialize activity statistics */
   GraphActivityInit ();

   /* initialize the language environment */
   if (VMSok (status = lib$get_users_language (&NaturalLanguageDsc)))
   {
      for (cptr = NaturalLanguage; *cptr && !ISLWS(*cptr); cptr++);
      *cptr = '\0';
      if (strsame (NaturalLanguage, "ENGLISH"))
         NaturalLanguageEnglish = true;
      else
      {
         NaturalLanguageEnglish = false;
         fprintf (stdout, "%%%s-I-LANGUAGE, natural language is %s\n",
                  Utility, NaturalLanguage);
      }
   }
   else
   {
      if (status == LIB$_ENGLUSED)
      {
         NaturalLanguageEnglish = true;
         strcpy (NaturalLanguage, "ENGLISH");
         fprintf (stdout,
            "%%%s-W-LANGUAGE, natural language has defaulted to %s\n",
            Utility, NaturalLanguage);
      }
      else
         ErrorExitVmsStatus (status, "lib$get_users_language()", FI_LI);
   }

   /* set up to allow the HTTPD process to be controlled */
   if (VMSnok (status = ControlHttpd ()))
      ErrorExitVmsStatus (status, "ControlHttpd()", FI_LI);

   /* initialize request history mechanism (ensure it's reasonable!) */
   RequestHistoryMax = Config.RequestHistory;
   if (RequestHistoryMax > 999) RequestHistoryMax = 0;

   /* disable process swapping */
   if (NoSwapOut)
   {
      if (VMSnok (status = sys$setprv (1, &PswapmMask, 0, 0)))
         ErrorExitVmsStatus (status, "sys$setprv()", FI_LI);
      if (VMSnok (status = sys$setswm (1)))
         ErrorExitVmsStatus (status, "sys$setswm()", FI_LI);
      if (VMSnok (status = sys$setprv (0, &PswapmMask, 0, 0)))
         ErrorExitVmsStatus (status, "sys$setprv()", FI_LI);
   }

   /* set up and declare the exit handler */
   ExitHandler.HandlerAddress = &HttpdExit;
   ExitHandler.ArgCount = 1;
   ExitHandler.ExitStatusPtr = &ExitStatus;
   if (VMSnok (status = sys$dclexh (&ExitHandler)))
      ErrorExitVmsStatus (status, "sys$dclexh()", FI_LI);

   /* disable AST's until we've created the listening sockets */
   sys$setast (0);

   /* need SYSPRV to bind to a well-known port */
   EnableSysPrv();

   /* create network services */
   NetCreateService ();

   /* turn off SYSPRV after binding to well-known port */
   DisableSysPrv();

   /* initialize logging after service creation (in case of per-service logs) */
   if (VMSnok (status = Logging (NULL, LOGGING_BEGIN)))
      ErrorExitVmsStatus (status, "Logging()", FI_LI);

   /* ensure any alterations to accounting during startup show up */
   if (MonitorEnabled)
      if (VMSnok (status = DefineMonitorLogicals (NULL)))
         ErrorExitVmsStatus (status, "DefineMonitorLogicals()", FI_LI);

   /*****************************/
   /* begin to process requests */
   /*****************************/

   /* reenable user-mode ASTs to allow use of our services */
   sys$setast (1);

   sys$hiber ();

   exit (SS$_NORMAL);
}

/*****************************************************************************/
/*
If we can't get SYSPRV when we need it, or more importantly, if we can't turn
it off when we're finished with it the exit the server!
*/ 

EnableSysPrv ()

{
   int  status;

   /*********/
   /* begin */
   /*********/

   if (Debug) fprintf (stdout, "EnableSysPrv()\n");

#if OPERATE_WITH_SYSPRV
   if (OperateWithSysPrv) return;
#endif

   if (VMSok (status = sys$setprv (1, &SysPrvMask, 0, 0))) return;
   ErrorExitVmsStatus (status, "sys$setprv()", FI_LI);
}

/*****************************************************************************/
/*
If we can't get SYSPRV when we need it, or more importantly, if we can't turn
it off when we're finished with it the exit the server!
*/ 

DisableSysPrv ()

{
   int  status;

   /*********/
   /* begin */
   /*********/

   if (Debug) fprintf (stdout, "DisableSysPrv()\n");

#if OPERATE_WITH_SYSPRV
   if (OperateWithSysPrv) return;
#endif

   if (VMSok (status = sys$setprv (0, &SysPrvMask, 0, 0))) return;
   ErrorExitVmsStatus (status, "sys$setprv()", FI_LI);
}

/*****************************************************************************/
/*
Exit the HTTP server, via this declared exit handler.
*/

HttpdExit (unsigned long *ExitStatusPtr)

{
   if (Debug) fprintf (stdout, "HttpdExit() %%X%08.08X\n", *ExitStatusPtr);

   /* run down any subprocesses associated with script processing */
   DclExit ();

   if (LoggingEnabled)
   {
      /* provide a server exit entry */
      Logging (NULL, LOGGING_END);
   }

   if (MonitorEnabled)
   {
      /* set exit status in logical names used by HTTPd monitor utility */
      Accounting.LastExitStatus = *ExitStatusPtr;
      DefineMonitorLogicals (NULL);
   }
}

/*****************************************************************************/
/*
As the code gets more complex it's becoming increasingly possible a coding or
design error will leave privileges turned on somewhere. To help detect any such
problem ensure everthing's off occasionlly (when there are no more connections
to process).
*/

#ifdef __DECC
#pragma inline(HttpdCheckPriv)
#endif

HttpdCheckPriv ()

{
   static long  Pid = 0;
   static unsigned long  JpiCurPriv [2];
   static struct
   {
      unsigned short  buf_len;
      unsigned short  item;
      unsigned char   *buf_addr;
      unsigned short  *ret_len;
   }
      JpiItems [] =
   {
      { sizeof(JpiCurPriv), JPI$_CURPRIV, &JpiCurPriv, 0 },
      {0,0,0,0}
   };

   int  status;

   /*********/
   /* begin */
   /*********/

   if (Debug) fprintf (stdout, "HttpdCheckPriv()\n");

   if (VMSnok (status = sys$getjpiw (0, &Pid, 0, &JpiItems, 0, 0, 0)))
      ErrorExitVmsStatus (status, "sys$getjpi()", FI_LI);

   if ((JpiCurPriv[0] & ~AveJoePrvMask[0]) || JpiCurPriv[1])
   {
#if OPERATE_WITH_SYSPRV
      if (OperateWithSysPrv && (JpiCurPriv[0] & ~SysPrvMask[0])) return;
#else
      if (!(JpiCurPriv[0] & ~SysPrvMask[0])) return;
#endif

      if (Debug)
         fprintf (stdout, "%08.08X%08.08X\n", JpiCurPriv[1], JpiCurPriv[0]);
      ErrorExitVmsStatus (0, ErrorSanityCheck, FI_LI);
   }
}

/*****************************************************************************/
/*
Set the timer counters for the request according to the function code. 
HttpdSupervisor() will monitor these counters.  Due to the way the seconds are
counted-down, and the setting of a timer may be in the middle of any one
second interval, increment very small periods to ensure they are at least that.
*/

HttpdTimerSet
(
struct RequestStruct *rqptr,
int Function
)
{
   static boolean  Initialize = true;
   static unsigned long  InputSeconds,
                         KeepAliveSeconds,
                         NoProgressSeconds,
                         OutputSeconds;

   int  status;

   /*********/
   /* begin */
   /*********/

   if (Debug) fprintf (stdout, "HttpdTimerSet() %d\n", Function);

   if (Initialize)
   {
      Initialize = false;
      if (Config.TimeoutMinutesInput)
         InputSeconds = Config.TimeoutMinutesInput * 60;
      else
         InputSeconds = DEFAULT_TIMEOUT_INPUT_MINUTES * 60;
      if (Config.TimeoutMinutesOutput)
         OutputSeconds = Config.TimeoutMinutesOutput * 60;
      else
         OutputSeconds = DEFAULT_TIMEOUT_OUTPUT_MINUTES * 60;
      if (Config.TimeoutMinutesNoProgress)
         NoProgressSeconds = Config.TimeoutMinutesNoProgress * 60;
      else
         NoProgressSeconds = DEFAULT_TIMEOUT_NOPROGRESS_MINUTES * 60;
      if (Config.TimeoutSecondsKeepAlive)
         KeepAliveSeconds = Config.TimeoutSecondsKeepAlive;
      else
         KeepAliveSeconds = DEFAULT_TIMEOUT_KEEPALIVE_SECONDS;
      /* ensure small periods are at least that duration */
      if (KeepAliveSeconds && KeepAliveSeconds < 10) KeepAliveSeconds++;
   }

   switch (Function)
   {
      case TIMER_INPUT :

         rqptr->TimerOutputCount =
            rqptr->TimerNoProgressBytesTx =
            rqptr->TimerNoProgressCount =
            rqptr->TimerNoProgressSeconds =
            rqptr->TimerKeepAliveCount = 0;
         rqptr->TimerInputCount = InputSeconds;

         if (rqptr->WatchItem &&
            (WatchEnabled & WATCH_TIMER))
            WatchThis (rqptr, FI_LI, WATCH_TIMER,
                       "INPUT !UL seconds", InputSeconds);

         return;

      case TIMER_OUTPUT :

         rqptr->TimerInputCount =
            rqptr->TimerNoProgressBytesTx =
            rqptr->TimerKeepAliveCount = 0;
         rqptr->TimerOutputCount = OutputSeconds;
         rqptr->TimerNoProgressCount =
            rqptr->TimerNoProgressSeconds = NoProgressSeconds;

         if (rqptr->WatchItem &&
            (WatchEnabled & WATCH_TIMER))
            WatchThis (rqptr, FI_LI, WATCH_TIMER,
                       "OUTPUT !UL seconds", OutputSeconds);

         return;

      case TIMER_KEEPALIVE :

         rqptr->TimerInputCount =
            rqptr->TimerOutputCount =
            rqptr->TimerNoProgressBytesTx =
            rqptr->TimerNoProgressCount =
            rqptr->TimerNoProgressSeconds = 0;
         rqptr->TimerKeepAliveCount = KeepAliveSeconds;

         if (rqptr->WatchItem &&
            (WatchEnabled & WATCH_TIMER))
            WatchThis (rqptr, FI_LI, WATCH_TIMER,
                       "KEEP-ALIVE !UL seconds", KeepAliveSeconds);

         return;

      default :
         ErrorExitVmsStatus (0, ErrorSanityCheck, FI_LI);
   }

   rqptr->TimerTerminatedCount = 0;
}

/*****************************************************************************/
/*
Whenever at least one request is being processed this function is called every
second to supervise the requests' progress.  Scanning the list of current
processes it decrements set timer counters.  When a counter reaches zero
the request is run down.

This function was introduced to address a suspected multi-thread timing problem
associated with setting and cancelling independent timers and ASTs for each
request.  It is probably no more expensive than setting and reseting VMS
timers for each request's processing, and does seem to give some measure (or
feeling of) more structure and control!
*/

HttpdSupervisor (boolean TimerExpired)

{
   static unsigned long  OneSecondDelta [2] = { -10000000, -1 };
   static boolean  TimerSet = false;

   register struct RequestStruct  *rqeptr;
   register struct ListEntryStruct  *leptr;

   int  status;
   char  *TimerTypePtr;

   /*********/
   /* begin */
   /*********/

   if (Debug)
      fprintf (stdout, "HttpdSupervisor() %d HeadPtr: %d\n",
               TimerExpired, RequestList.HeadPtr);

   if (!TimerExpired)
   {
      /* new request, if the supervisor is already running just return */
      if (TimerSet) return;

      /* kick the request supervisor timer into life */
      if (VMSnok (status =
          sys$setimr (0, &OneSecondDelta, &HttpdSupervisor, true, 0)))
         ErrorExitVmsStatus (status, "sys$setimr()", FI_LI);
      TimerSet = true;

      return;
   }

   /* must have got here via the supervisor's timer expiring */
   TimerSet = false;

   /* if nothing in the current request list return without sys$setimr() */
   if ((leptr = RequestList.HeadPtr) == NULL) return;

   /* process the current request list entries */
   for (/* from above */; leptr != NULL; leptr = leptr->NextPtr)
   {
      rqeptr = (struct RequestStruct*)leptr;

      if (Debug)
         fprintf (stdout,
"%d <- %d -> %d (%d) in: %d out: %d alive: %d terminated: %d\n",
                  leptr->PrevPtr, leptr, leptr->NextPtr, rqeptr,
                  rqeptr->TimerInputCount, rqeptr->TimerOutputCount,
                  rqeptr->TimerKeepAliveCount, rqeptr->TimerTerminatedCount);


      /* ordered in the most common occurance of timings */
      if (rqeptr->TimerOutputCount)
      {
         if (rqeptr->TimerNoProgressBytesTx == rqeptr->BytesTx)
         {
            /* note the start of any no-progress timing */
            if (rqeptr->WatchItem &&
                (WatchEnabled & WATCH_TIMER) &&
                rqeptr->TimerNoProgressCount == rqeptr->TimerNoProgressSeconds)
               WatchThis (rqeptr, FI_LI, WATCH_TIMER,
                          "NO-PROGRESS !UL seconds",
                          rqeptr->TimerNoProgressSeconds);

            if (rqeptr->TimerNoProgressCount) rqeptr->TimerNoProgressCount--;
         }
         else
         {
            rqeptr->TimerNoProgressCount = rqeptr->TimerNoProgressSeconds;
            rqeptr->TimerNoProgressBytesTx = rqeptr->BytesTx;
         }

         if (rqeptr->TimerNoProgressCount)
         {
            if (--rqeptr->TimerOutputCount)
               continue;
            else
               TimerTypePtr = "OUTPUT";
         }
         else
            TimerTypePtr = "NO-PROGRESS";
      }
      else
      if (rqeptr->TimerKeepAliveCount)
      {
         if (--rqeptr->TimerKeepAliveCount)
            continue;
         else
         {
            TimerTypePtr = "KEEP-ALIVE";
            if (rqeptr->ClientChannel) NetCloseSocket (rqeptr);
         }
      }
      else
      if (rqeptr->TimerInputCount)
      {
         if (--rqeptr->TimerInputCount)
            continue;
         else
         {
            TimerTypePtr = "INPUT";
            if (rqeptr->ClientChannel) NetCloseSocket (rqeptr);
         }
      }

      if (!rqeptr->TimerTerminatedCount++)
      {
         if (rqeptr->WatchItem &&
             (WatchEnabled & WATCH_TIMER))
            WatchThis (rqeptr, FI_LI, WATCH_TIMER,
                       "!AZ expired", TimerTypePtr);

         if (rqeptr->DclTaskPtr != NULL)
         {
            /* DCL involved, do any DCL-specific cleanup */
            DclConcludeTask (rqeptr->DclTaskPtr, true);
         }
         else
         if (rqeptr->DECnetTaskPtr != NULL)
         {
            /* DECnet (CGI or OSU) involved */
            DECnetDisconnectTask (rqeptr->DECnetTaskPtr);
         }
         else
         if (rqeptr->ProxyTaskPtr != NULL)
         {
            if (rqeptr->ProxyTaskPtr->ProxyChannel)
               ProxyCloseSocket (rqeptr->ProxyTaskPtr);
         }

         if (rqeptr->ClientChannel) NetCloseSocket (rqeptr);
      }
   }

   if (VMSnok (status =
       sys$setimr (0, &OneSecondDelta, &HttpdSupervisor, true, 0)))
      ErrorExitVmsStatus (status, "sys$setimr()", FI_LI);

   TimerSet = true;
}

/*****************************************************************************/

